"""Differentiable utils typically used when computing loss functions."""
import torch
import torch.nn.functional as F


def compute_advantages(discount, gae_lambda, max_path_length, baselines,
                       rewards):
    """Calculate advantages.

    Advantages are a discounted cumulative sum.

    Calculate advantages using a baseline (value function) according to
    Generalized Advantage Estimation (GAE)

    The discounted cumulative sum can be computed using conv2d with filter.
    filter:
        [1, (discount * gae_lambda), (discount * gae_lambda) ^ 2, ...]
        where the length is same with max_path_length.

    baselines and rewards are also has same shape.
        baselines:
        [ [b_11, b_12, b_13, ... b_1n],
          [b_21, b_22, b_23, ... b_2n],
          ...
          [b_m1, b_m2, b_m3, ... b_mn] ]
        rewards:
        [ [r_11, r_12, r_13, ... r_1n],
          [r_21, r_22, r_23, ... r_2n],
          ...
          [r_m1, r_m2, r_m3, ... r_mn] ]

    Args:
        discount (float): RL discount factor (i.e. gamma).
        gae_lambda (float): Lambda, as used for Generalized Advantage
            Estimation (GAE).
        max_path_length (int): Maximum length of a single rollout.
        baselines (torch.Tensor): A 2D vector of value function estimates with
            shape (N, T), where N is the batch dimension (number of episodes)
            and T is the maximum path length experienced by the agent. If an
            episode terminates in fewer than T time steps, the remaining
            elements in that episode should be set to 0.
        rewards (torch.Tensor): A 2D vector of per-step rewards with shape
            (N, T), where N is the batch dimension (number of episodes) and T
            is the maximum path length experienced by the agent. If an episode
            terminates in fewer than T time steps, the remaining elements in
            that episode should be set to 0.

    Returns:
        torch.Tensor: A 2D vector of calculated advantage values with shape
            (N, T), where N is the batch dimension (number of episodes) and T
            is the maximum path length experienced by the agent. If an episode
            terminates in fewer than T time steps, the remaining values in that
            episode should be set to 0.

    """
    adv_filter = torch.full((1, 1, 1, max_path_length - 1),
                            discount * gae_lambda)
    adv_filter = torch.cumprod(F.pad(adv_filter, (1, 0), value=1), dim=-1)

    deltas = (rewards + discount * F.pad(baselines, (0, 1))[:, 1:] - baselines)
    deltas = F.pad(deltas, (0, max_path_length - 1)).unsqueeze(0).unsqueeze(0)

    advantages = F.conv2d(deltas, adv_filter, stride=1).squeeze()
    return advantages



from tests.fixtures.envs.dummy.base import DummyEnv
from tests.fixtures.envs.dummy.dummy_box_env import DummyBoxEnv
from tests.fixtures.envs.dummy.dummy_dict_env import DummyDictEnv
from tests.fixtures.envs.dummy.dummy_discrete_2d_env import DummyDiscrete2DEnv
from tests.fixtures.envs.dummy.dummy_discrete_env import DummyDiscreteEnv
from tests.fixtures.envs.dummy.dummy_discrete_pixel_env import (
    DummyDiscretePixelEnv)
from tests.fixtures.envs.dummy.dummy_discrete_pixel_env_baselines import (
    DummyDiscretePixelEnvBaselines)
from tests.fixtures.envs.dummy.dummy_reward_box_env import DummyRewardBoxEnv

__all__ = [
    'DummyEnv', 'DummyBoxEnv', 'DummyDictEnv', 'DummyDiscrete2DEnv',
    'DummyDiscreteEnv', 'DummyDiscretePixelEnv',
    'DummyDiscretePixelEnvBaselines', 'DummyRewardBoxEnv'
]



"""A progress bar counter class."""
import dowel
from dowel import logger
import pyprind


class ProgBarCounter:
    """Record progress in a bar counter format."""

    def __init__(self, total_count):
        """Init function."""
        self.total_count = total_count
        self.max_progress = 1000000
        self.cur_progress = 0
        self.cur_count = 0
        if logger.has_output_type(dowel.StdOutput):
            self.pbar = pyprind.ProgBar(self.max_progress)
        else:
            self.pbar = None

    def inc(self, increment):
        """Increment function."""
        if logger.has_output_type(dowel.StdOutput):
            self.cur_count += increment
            new_progress = (
                self.cur_count * self.max_progress / self.total_count)
            if new_progress < self.max_progress:
                self.pbar.update(new_progress - self.cur_progress)
            self.cur_progress = new_progress

    def stop(self):
        """Stop function."""
        if self.pbar is not None and self.pbar.active:
            self.pbar.stop()



import numpy as np

from garage.envs.wrappers import Noop
from tests.fixtures.envs.dummy import DummyDiscretePixelEnv


class TestNoop:
    def test_noop(self):
        env = Noop(DummyDiscretePixelEnv(), noop_max=3)

        for _ in range(1000):
            env.reset()
            assert 1 <= env.env.step_called <= 3

        env = Noop(DummyDiscretePixelEnv(), noop_max=10)
        for _ in range(1000):
            obs = env.reset()
            if env.env.step_called % 5 == 0:
                """
                There are only 5 lives in the environment, so if number of
                steps are multiple of 5, env will call reset at last.
                """
                assert np.array_equal(obs,
                                      np.ones(env.observation_space.shape))
            else:
                assert not np.array_equal(obs,
                                          np.ones(env.observation_space.shape))



"""Create a test that fails when garage.tf.algos.TD3 performance is too low."""

import gym
import tensorflow as tf

from garage.np.exploration_strategies.gaussian_strategy import GaussianStrategy
from garage.replay_buffer import SimpleReplayBuffer
from garage.tf.algos import TD3
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import ContinuousMLPPolicy
from garage.tf.q_functions import ContinuousMLPQFunction
from tests.fixtures import snapshot_config, TfGraphTestCase


class TestTD3(TfGraphTestCase):
    """Tests for TD3 algo."""

    def test_td3_pendulum(self):
        """Test TD3 with Pendulum environment."""
        with LocalTFRunner(snapshot_config) as runner:
            env = TfEnv(gym.make('InvertedDoublePendulum-v2'))

            action_noise = GaussianStrategy(env.spec,
                                            max_sigma=0.1,
                                            min_sigma=0.1)

            policy = ContinuousMLPPolicy(env_spec=env.spec,
                                         hidden_sizes=[400, 300],
                                         hidden_nonlinearity=tf.nn.relu,
                                         output_nonlinearity=tf.nn.tanh)

            qf = ContinuousMLPQFunction(name='ContinuousMLPQFunction',
                                        env_spec=env.spec,
                                        hidden_sizes=[400, 300],
                                        action_merge_layer=0,
                                        hidden_nonlinearity=tf.nn.relu)

            qf2 = ContinuousMLPQFunction(name='ContinuousMLPQFunction2',
                                         env_spec=env.spec,
                                         hidden_sizes=[400, 300],
                                         action_merge_layer=0,
                                         hidden_nonlinearity=tf.nn.relu)

            replay_buffer = SimpleReplayBuffer(env_spec=env.spec,
                                               size_in_transitions=int(1e6),
                                               time_horizon=250)

            algo = TD3(env_spec=env.spec,
                       policy=policy,
                       policy_lr=1e-3,
                       qf_lr=1e-3,
                       qf=qf,
                       qf2=qf2,
                       replay_buffer=replay_buffer,
                       target_update_tau=0.005,
                       n_epoch_cycles=20,
                       n_train_steps=50,
                       discount=0.99,
                       smooth_return=False,
                       min_buffer_size=int(1e4),
                       buffer_batch_size=100,
                       policy_weight_decay=0.001,
                       qf_weight_decay=0.001,
                       exploration_strategy=action_noise,
                       policy_optimizer=tf.compat.v1.train.AdamOptimizer,
                       qf_optimizer=tf.compat.v1.train.AdamOptimizer)

            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10,
                                        n_epoch_cycles=20,
                                        batch_size=250)
            assert last_avg_ret > 400



import tensorflow as tf

from examples.AV.example_runner_ga_av import runner as ga_runner


def validate_ga():
    # Overall settings
    max_path_length = 50
    s_0 = [0.0, -4.0, 1.0, 11.17, -35.0]
    base_log_dir = './data'
    # experiment settings
    run_experiment_args = {'snapshot_mode': 'last',
                           'snapshot_gap': 1,
                           'log_dir': None,
                           'exp_name': None,
                           'seed': 0,
                           'n_parallel': 1,
                           'tabular_log_file': 'progress.csv'
                           }

    # runner settings
    runner_args = {'n_epochs': 2,
                   'batch_size': 500,
                   'plot': False
                   }

    # env settings
    env_args = {'id': 'ast_toolbox:GoExploreAST-v1',
                'blackbox_sim_state': False,
                'open_loop': False,
                'fixed_init_state': True,
                's_0': s_0,
                }

    # simulation settings
    sim_args = {'blackbox_sim_state': False,
                'open_loop': False,
                'fixed_initial_state': True,
                'max_path_length': max_path_length
                }

    # reward settings
    reward_args = {'use_heuristic': True}

    # spaces settings
    spaces_args = {'open_loop': False}

    # policy settings
    policy_args = {
        'hidden_sizes': (64, 32),
        'output_nonlinearity': tf.nn.tanh,
    }

    sampler_args = {'n_envs': 1,
                    'open_loop': False}

    # GA Settings

    ga_bpq_args = {'N': 10}

    exp_log_dir = base_log_dir
    max_path_length = 50
    s_0 = [0.0, -4.0, 1.0, 11.17, -35.0]
    env_args['s_0'] = s_0
    reward_args['use_heuristic'] = True
    sim_args['max_path_length'] = max_path_length

    for ga_type in ['ga', 'gasm']:
        run_experiment_args['log_dir'] = exp_log_dir + '/' + ga_type
        run_experiment_args['exp_name'] = ga_type
        ga_algo_args = {'max_path_length': max_path_length,
                        'batch_size': 100,
                        'pop_size': 5,
                        'truncation_size': 3,
                        'keep_best': 1,
                        'step_size': 0.01,
                        'n_itr': 2,
                        'log_interval': 100,
                        }
        ga_runner(
            ga_type=ga_type,
            env_args=env_args,
            run_experiment_args=run_experiment_args,
            sim_args=sim_args,
            reward_args=reward_args,
            spaces_args=spaces_args,
            policy_args=policy_args,
            algo_args=ga_algo_args,
            bpq_args=ga_bpq_args,
            sampler_args=sampler_args,
            runner_args=runner_args,
        )

    return True


if __name__ == '__main__':
    validate_ga()



# Import the example classes
from garage.envs.normalized_env import normalize
from garage.experiment import run_experiment
from garage.np.baselines.linear_feature_baseline import LinearFeatureBaseline
# Import the necessary garage classes
from garage.tf.algos.trpo import TRPO
from garage.tf.envs.base import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies.gaussian_lstm_policy import GaussianLSTMPolicy

# Import the AST classes
from ast_toolbox.envs import ASTEnv
from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.samplers import ASTVectorizedSampler
from ast_toolbox.simulators import ExampleAVSimulator
from ast_toolbox.spaces import ExampleAVSpaces

# Useful imports

batch_size = 4000
max_path_length = 50
n_envs = batch_size // max_path_length


def run_task(snapshot_config, *_):
    with LocalTFRunner(
            snapshot_config=snapshot_config, max_cpus=2) as runner:
        # Instantiate the example classes
        sim = ExampleAVSimulator()
        reward_function = ExampleAVReward()
        spaces = ExampleAVSpaces()

        # Create the environment
        env = TfEnv(normalize(ASTEnv(blackbox_sim_state=True,
                                     fixed_init_state=True,
                                     s_0=[-0.5, -4.0, 1.0, 11.17, -35.0],
                                     simulator=sim,
                                     reward_function=reward_function,
                                     spaces=spaces
                                     )))

        # Instantiate the garage objects
        policy = GaussianLSTMPolicy(name='lstm_policy',
                                    env_spec=env.spec,
                                    hidden_dim=64)

        baseline = LinearFeatureBaseline(env_spec=env.spec)

        algo = TRPO(
            env_spec=env.spec,
            policy=policy,
            baseline=baseline,
            max_path_length=max_path_length,
            discount=0.99,
            kl_constraint='soft',
            max_kl_step=0.01)

        sampler_cls = ASTVectorizedSampler

        runner.setup(
            algo=algo,
            env=env,
            sampler_cls=sampler_cls,
            sampler_args={"sim": sim,
                          "reward_function": reward_function})

        runner.train(n_epochs=1, batch_size=4000, plot=False)

        print("Installation successfully validated")


def validate_parallel():
    run_experiment(run_task, snapshot_mode='last', seed=1, n_parallel=2, log_dir='./data/parallel')
    return True


if __name__ == '__main__':
    validate_parallel()



import numpy as np
import tensorflow as tf

from garage.tf.models.parameter import parameter
from garage.tf.models.parameter import recurrent_parameter
from tests.fixtures import TfGraphTestCase


class TestParameter(TfGraphTestCase):

    def setup_method(self):
        super().setup_method()
        self.input_vars = tf.placeholder(shape=[None, 2, 5], dtype=tf.float32)
        self.step_input_vars = tf.placeholder(shape=[None, 5],
                                              dtype=tf.float32)
        self.initial_params = np.array([48, 21, 33])

        self.data = np.zeros(shape=[5, 2, 5])
        self.step_data = np.zeros(shape=[5, 5])
        self.feed_dict = {
            self.input_vars: self.data,
            self.step_input_vars: self.step_data
        }

    def test_param(self):
        param = parameter(input_var=self.input_vars,
                          length=3,
                          initializer=tf.constant_initializer(
                              self.initial_params))
        self.sess.run(tf.global_variables_initializer())
        p = self.sess.run(param, feed_dict=self.feed_dict)

        assert p.shape == (5, 3)
        assert np.all(p == self.initial_params)

    def test_recurrent_param(self):
        param, _ = recurrent_parameter(input_var=self.input_vars,
                                       step_input_var=self.step_input_vars,
                                       length=3,
                                       initializer=tf.constant_initializer(
                                           self.initial_params))
        self.sess.run(tf.compat.v1.global_variables_initializer())
        p = self.sess.run(param, feed_dict=self.feed_dict)

        assert p.shape == (5, 2, 3)
        assert np.array_equal(p, np.full([5, 2, 3], self.initial_params))



"""Environments for formulating validation as an AST reinforcement learning problem."""
import importlib
__all__ = []
if importlib.util.find_spec('bsddb3') is not None:
    # Only load bsddb3 dependent packages if it is installed
    from ast_toolbox.envs.go_explore_ast_env import GoExploreASTEnv  # noqa
    from ast_toolbox.envs.go_explore_ast_env import Custom_GoExploreASTEnv  # noqa

    __all__ += ['GoExploreASTEnv', 'Custom_GoExploreASTEnv']

from .ast_env import ASTEnv  # noqa

__all__ += ['ASTEnv']



"""Dummy distribution for testing purpose."""
from garage.tf.distributions.base import Distribution


class DummyDistribution(Distribution):
    """Dummy distribution for testing purpose."""

    @property
    def dim(self):
        """Distribution dimension.

        Returns:
            int: Distribution dimenison.

        """
        return 1

    def kl_sym(self, old_dist_info_vars, new_dist_info_vars):
        """Compute the symbolic KL divergence of two distributions.

        Args:
            old_dist_info_vars (dict): Old distribution tensors.
            new_dist_info_vars (dict): New distribution tensors.
        """

    def kl(self, old_dist_info, new_dist_info):
        """Compute the KL divergence of two distributions.

        Args:
            old_dist_info (dict): Old distribution parameters.
            new_dist_info (dict): New distribution parameters.
        """

    def likelihood_ratio_sym(self, x_var, old_dist_info_vars,
                             new_dist_info_vars):
        """Likelihood ratio sym.

        Args:
            x_var (tf.Tensor): Input placeholder.
            old_dist_info_vars (dict): Old distribution tensors.
            new_dist_info_vars (dict): New distribution tensors.
        """

    def entropy(self, dist_info):
        """Entropy.

        Args:
            dist_info (dict): Distribution parameters.

        """

    def log_likelihood_sym(self, x_var, dist_info_vars):
        """Log Likelihood sym.

        Args:
            x_var (tf.Tensor): Input placeholder.
            dist_info_vars (dict): Distribution tensors.

        """

    def log_likelihood(self, xs, dist_info):
        """Log Likelihood.

        Args:
            xs (np.ndarray): Inputs.
            dist_info (dict): Distribution parameters.

        """

    @property
    def dist_info_specs(self):
        """Distribution information specification.

        Returns:
            list[tuple]: Key for distribution information and shape.

        """
        return [('dummy', (1, ))]



import numpy as np
import pytest
import tensorflow as tf

from garage.tf.models.mlp import mlp
from tests.fixtures import TfGraphTestCase


class TestMLPConcat(TfGraphTestCase):

    def setup_method(self):
        super(TestMLPConcat, self).setup_method()
        self.obs_input = np.array([[1, 2, 3, 4]])
        self.act_input = np.array([[1, 2, 3, 4]])
        input_shape_1 = self.obs_input.shape[1:]  # 4
        input_shape_2 = self.act_input.shape[1:]  # 4
        self.hidden_nonlinearity = tf.nn.relu

        self._obs_input = tf.compat.v1.placeholder(tf.float32,
                                                   shape=(None, ) +
                                                   input_shape_1,
                                                   name='input')

        self._act_input = tf.compat.v1.placeholder(tf.float32,
                                                   shape=(None, ) +
                                                   input_shape_2,
                                                   name='input')

        self._output_shape = 2

        # We build a default mlp
        with tf.compat.v1.variable_scope('MLP_Concat'):
            self.mlp_f = mlp(input_var=self._obs_input,
                             output_dim=self._output_shape,
                             hidden_sizes=(32, 32),
                             input_var2=self._act_input,
                             concat_layer=0,
                             hidden_nonlinearity=self.hidden_nonlinearity,
                             name='mlp1')

        self.sess.run(tf.compat.v1.global_variables_initializer())

    def test_multiple_same_mlp(self):
        # We create another mlp with the same name, trying to reuse it
        with tf.compat.v1.variable_scope('MLP_Concat', reuse=True):
            self.mlp_same_copy = mlp(
                input_var=self._obs_input,
                output_dim=self._output_shape,
                hidden_sizes=(32, 32),
                input_var2=self._act_input,
                concat_layer=0,
                hidden_nonlinearity=self.hidden_nonlinearity,
                name='mlp1')

        # We modify the weight of the default mlp and feed
        # The another mlp created should output the same result
        with tf.compat.v1.variable_scope('MLP_Concat', reuse=True):
            w = tf.compat.v1.get_variable('mlp1/hidden_0/kernel')
            self.sess.run(w.assign(w + 1))
            mlp_output = self.sess.run(self.mlp_f,
                                       feed_dict={
                                           self._obs_input: self.obs_input,
                                           self._act_input: self.act_input
                                       })
            mlp_output2 = self.sess.run(self.mlp_same_copy,
                                        feed_dict={
                                            self._obs_input: self.obs_input,
                                            self._act_input: self.act_input
                                        })

        np.testing.assert_array_almost_equal(mlp_output, mlp_output2)

    def test_different_mlp(self):
        # We create another mlp with different name
        with tf.compat.v1.variable_scope('MLP_Concat'):
            self.mlp_different_copy = mlp(
                input_var=self._obs_input,
                output_dim=self._output_shape,
                hidden_sizes=(32, 32),
                input_var2=self._act_input,
                concat_layer=0,
                hidden_nonlinearity=self.hidden_nonlinearity,
                name='mlp2')

        # Initialize the new mlp variables
        self.sess.run(tf.compat.v1.global_variables_initializer())

        # We modify the weight of the default mlp and feed
        # The another mlp created should output different result
        with tf.compat.v1.variable_scope('MLP_Concat', reuse=True):
            w = tf.compat.v1.get_variable('mlp1/hidden_0/kernel')
            self.sess.run(w.assign(w + 1))
            mlp_output = self.sess.run(self.mlp_f,
                                       feed_dict={
                                           self._obs_input: self.obs_input,
                                           self._act_input: self.act_input
                                       })
            mlp_output2 = self.sess.run(self.mlp_different_copy,
                                        feed_dict={
                                            self._obs_input: self.obs_input,
                                            self._act_input: self.act_input
                                        })

        assert not np.array_equal(mlp_output, mlp_output2)

    def test_output_shape(self):
        mlp_output = self.sess.run(self.mlp_f,
                                   feed_dict={
                                       self._obs_input: self.obs_input,
                                       self._act_input: self.act_input
                                   })

        assert mlp_output.shape[1] == self._output_shape

    def test_output_value(self):
        with tf.compat.v1.variable_scope('MLP_Concat', reuse=True):
            h1_w = tf.compat.v1.get_variable('mlp1/hidden_0/kernel')
            h1_b = tf.compat.v1.get_variable('mlp1/hidden_0/bias')
            h2_w = tf.compat.v1.get_variable('mlp1/hidden_1/kernel')
            h2_b = tf.compat.v1.get_variable('mlp1/hidden_1/bias')
            out_w = tf.compat.v1.get_variable('mlp1/output/kernel')
            out_b = tf.compat.v1.get_variable('mlp1/output/bias')

        mlp_output = self.sess.run(self.mlp_f,
                                   feed_dict={
                                       self._obs_input: self.obs_input,
                                       self._act_input: self.act_input
                                   })

        # First layer
        h2_in = tf.matmul(tf.concat([self._obs_input, self._act_input], 1),
                          h1_w) + h1_b
        h2_in = self.hidden_nonlinearity(h2_in)

        # Second layer
        h3_in = tf.matmul(h2_in, h2_w) + h2_b
        h3_in = self.hidden_nonlinearity(h3_in)

        # Output layer
        h3_out = tf.matmul(h3_in, out_w) + out_b
        out = self.sess.run(h3_out,
                            feed_dict={
                                self._obs_input: self.obs_input,
                                self._act_input: self.act_input
                            })

        np.testing.assert_array_equal(out, mlp_output)

    def test_layer_normalization(self):
        # Create a mlp with layer normalization
        with tf.compat.v1.variable_scope('MLP_Concat'):
            self.mlp_f_w_n = mlp(input_var=self._obs_input,
                                 output_dim=self._output_shape,
                                 hidden_sizes=(32, 32),
                                 input_var2=self._act_input,
                                 concat_layer=0,
                                 hidden_nonlinearity=self.hidden_nonlinearity,
                                 name='mlp2',
                                 layer_normalization=True)

        # Initialize the new mlp variables
        self.sess.run(tf.compat.v1.global_variables_initializer())

        with tf.compat.v1.variable_scope('MLP_Concat', reuse=True):
            h1_w = tf.compat.v1.get_variable('mlp2/hidden_0/kernel')
            h1_b = tf.compat.v1.get_variable('mlp2/hidden_0/bias')
            h2_w = tf.compat.v1.get_variable('mlp2/hidden_1/kernel')
            h2_b = tf.compat.v1.get_variable('mlp2/hidden_1/bias')
            out_w = tf.compat.v1.get_variable('mlp2/output/kernel')
            out_b = tf.compat.v1.get_variable('mlp2/output/bias')
            beta_1 = tf.compat.v1.get_variable('mlp2/LayerNorm/beta')
            gamma_1 = tf.compat.v1.get_variable('mlp2/LayerNorm/gamma')
            beta_2 = tf.compat.v1.get_variable('mlp2/LayerNorm_1/beta')
            gamma_2 = tf.compat.v1.get_variable('mlp2/LayerNorm_1/gamma')

        # First layer
        y = tf.matmul(tf.concat([self._obs_input, self._act_input], 1),
                      h1_w) + h1_b
        y = self.hidden_nonlinearity(y)
        mean, variance = tf.nn.moments(y, [1], keep_dims=True)
        normalized_y = (y - mean) / tf.sqrt(variance + 1e-12)
        y_out = normalized_y * gamma_1 + beta_1

        # Second layer
        y = tf.matmul(y_out, h2_w) + h2_b
        y = self.hidden_nonlinearity(y)
        mean, variance = tf.nn.moments(y, [1], keep_dims=True)
        normalized_y = (y - mean) / tf.sqrt(variance + 1e-12)
        y_out = normalized_y * gamma_2 + beta_2

        # Output layer
        y = tf.matmul(y_out, out_w) + out_b

        out = self.sess.run(y,
                            feed_dict={
                                self._obs_input: self.obs_input,
                                self._act_input: self.act_input
                            })
        mlp_output = self.sess.run(self.mlp_f_w_n,
                                   feed_dict={
                                       self._obs_input: self.obs_input,
                                       self._act_input: self.act_input
                                   })

        np.testing.assert_array_almost_equal(out, mlp_output)

    @pytest.mark.parametrize('concat_idx', [2, 1, 0, -1, -2])
    def test_concat_layer(self, concat_idx):
        with tf.compat.v1.variable_scope('mlp_concat_test'):
            _ = mlp(input_var=self._obs_input,
                    output_dim=self._output_shape,
                    hidden_sizes=(64, 32),
                    input_var2=self._act_input,
                    concat_layer=concat_idx,
                    hidden_nonlinearity=self.hidden_nonlinearity,
                    name='mlp2')
        obs_input_size = self._obs_input.shape[1].value
        act_input_size = self._act_input.shape[1].value

        expected_units = [obs_input_size, 64, 32]
        expected_units[concat_idx] += act_input_size

        actual_units = []
        with tf.compat.v1.variable_scope('mlp_concat_test', reuse=True):
            h1_w = tf.compat.v1.get_variable('mlp2/hidden_0/kernel')
            h2_w = tf.compat.v1.get_variable('mlp2/hidden_1/kernel')
            out_w = tf.compat.v1.get_variable('mlp2/output/kernel')

            actual_units.append(h1_w.shape[0].value)
            actual_units.append(h2_w.shape[0].value)
            actual_units.append(out_w.shape[0].value)

        assert np.array_equal(expected_units, actual_units)

    @pytest.mark.parametrize('concat_idx', [2, 1, 0, -1, -2])
    def test_invalid_concat_args(self, concat_idx):
        with tf.compat.v1.variable_scope('mlp_concat_test'):
            _ = mlp(input_var=self._obs_input,
                    output_dim=self._output_shape,
                    hidden_sizes=(64, 32),
                    concat_layer=concat_idx,
                    hidden_nonlinearity=self.hidden_nonlinearity,
                    name='mlp_no_input2')

        obs_input_size = self._obs_input.shape[1].value

        # concat_layer argument should be silently ignored.
        expected_units = [obs_input_size, 64, 32]

        actual_units = []
        with tf.compat.v1.variable_scope('mlp_concat_test', reuse=True):
            h1_w = tf.compat.v1.get_variable('mlp_no_input2/hidden_0/kernel')
            h2_w = tf.compat.v1.get_variable('mlp_no_input2/hidden_1/kernel')
            out_w = tf.compat.v1.get_variable('mlp_no_input2/output/kernel')

            actual_units.append(h1_w.shape[0].value)
            actual_units.append(h2_w.shape[0].value)
            actual_units.append(out_w.shape[0].value)

        assert np.array_equal(expected_units, actual_units)

    @pytest.mark.parametrize('concat_idx', [2, 1, 0, -1, -2])
    def test_no_hidden(self, concat_idx):
        with tf.compat.v1.variable_scope('mlp_concat_test'):
            _ = mlp(input_var=self._obs_input,
                    output_dim=self._output_shape,
                    hidden_sizes=(),
                    input_var2=self._act_input,
                    concat_layer=concat_idx,
                    hidden_nonlinearity=self.hidden_nonlinearity,
                    name='mlp2')

        obs_input_size = self._obs_input.shape[1].value
        act_input_size = self._act_input.shape[1].value

        # concat_layer argument should be reset to point to input_var.
        expected_units = [obs_input_size]
        expected_units[0] += act_input_size

        actual_units = []
        with tf.compat.v1.variable_scope('mlp_concat_test', reuse=True):
            out_w = tf.compat.v1.get_variable('mlp2/output/kernel')
            actual_units.append(out_w.shape[0].value)

        assert np.array_equal(expected_units, actual_units)



import time

# def identity(*args):
#     if len(args) == 1:
#         return args[0]
#     return args


class TransitionModel:
    """The wrapper for the transitin model used in the tree search.

    Parameters
    ----------
    getInitialState : function
        getInitialState() returns the initial AST state.
    getNextState : function
        getNextState(s, a) returns the next state and the reward.
    isEndState : function
        isEndState(s) returns whether s is a terminal state.
    maxSteps : int
        The maximum path length.
    goToState : function
        goToState(s) sets the simulator to the target state s.
    """

    def __init__(self, getInitialState, getNextState, isEndState, maxSteps, goToState):
        self.getInitialState = getInitialState
        self.getNextState = getNextState
        self.isEndState = isEndState
        self.maxSteps = maxSteps
        self.goToState = goToState


def simulate(model, p, policy, verbose=False, sleeptime=0.0):
    """Simulate the environment model using the policy and the parameter p.

    Parameters
    ----------
    model : :py:class:`ast_toolbox.mcts.MDP.TransitionModel`
        The environment model.
    p :
        The extra paramters needed by the policy.
    policy : function
        policy(p, s) returns the next action.
    verbose : bool, optional
        Whether to logging simulating information.
    sleeptime: float, optional
        The pause time between each step.

    Returns
    ----------
    cum_reward : float
        The cumulative reward.
    actions: list
        The action sequence of the path.
    """
    cum_reward = 0.0
    actions = []
    s = model.getInitialState()
    for i in range(model.maxSteps):
        # if verbose:
        # 	print("Step: ",i," of ", model.maxSteps)
        a = policy(p, s)
        actions.append(a)
        s, r = model.getNextState(s, a)
        time.sleep(sleeptime)
        cum_reward += r
        if model.isEndState(s):
            break
    if verbose:
        print("End at: ", i, " of ", model.maxSteps)
    return cum_reward, actions



import numpy as np
from dowel import logger
from garage.tf.misc import tensor_utils
from garage.tf.optimizers.conjugate_gradient_optimizer import PearlmutterHvp
from garage.tf.optimizers.utils import LazyDict
from garage.tf.optimizers.utils import sliced_fun


class DirectionConstraintOptimizer:
    """Performs constrained optimization via line search on the given gradient direction.

    Parameters
    ----------
    cg_iters : int, optional
        The number of CG iterations used to calculate A^-1 g.
    reg_coeff : float, optional
        A small value so that A -> A + reg*I.
    subsample_factor : int, optional
        Subsampling factor to reduce samples when using "conjugate gradient. Since the
        computation time for the descent direction dominates, this can greatly reduce the overall computation time.
    debug_nan : bool, optional
        If set to True, NanGuard will be added to the compilation, and ipdb will be invoked when
        nan is detected.
    accept_violation : bool, optional
        Whether to accept the descent step if it violates the line search condition after
        exhausting all backtracking budgets.
    """

    def __init__(
            self,
            cg_iters=10,
            reg_coeff=1e-5,
            subsample_factor=1.,
            backtrack_ratio=0.8,
            max_backtracks=15,
            debug_nan=False,
            accept_violation=False,
            hvp_approach=None,
            num_slices=1):
        self._cg_iters = cg_iters
        self._reg_coeff = reg_coeff
        self._subsample_factor = subsample_factor
        self._backtrack_ratio = backtrack_ratio
        self._max_backtracks = max_backtracks
        self._num_slices = num_slices

        self._opt_fun = None
        self._target = None
        self._max_constraint_val = None
        self._constraint_name = None
        self._debug_nan = debug_nan
        self._accept_violation = accept_violation
        if hvp_approach is None:
            hvp_approach = PearlmutterHvp(num_slices)
        self._hvp_approach = hvp_approach

    def update_opt(self, target, leq_constraint, inputs, extra_inputs=None, constraint_name="constraint", *args,
                   **kwargs):
        """Update the internal tensowflow operations.

        Parameters
        ----------
        target :
            A parameterized object to optimize over. It should implement methods of the
            :py:class:`garage.core.paramerized.Parameterized` class.
        leq_constraint : :py:class:'tensorflow.Tensor'
            The variable to be constrained.
        inputs :
            A list of symbolic variables as inputs, which could be subsampled if needed. It is assumed
            that the first dimension of these inputs should correspond to the number of data points.
        extra_inputs :
            A list of symbolic variables as extra inputs which should not be subsampled.
        """

        inputs = tuple(inputs)
        if extra_inputs is None:
            extra_inputs = tuple()
        else:
            extra_inputs = tuple(extra_inputs)

        # constraint_term, constraint_value = leq_constraint
        constraint_term = leq_constraint

        # params = target.get_params(trainable=True)

        self._hvp_approach.update_hvp(f=constraint_term, target=target, inputs=inputs + extra_inputs,
                                      reg_coeff=self._reg_coeff)

        self._target = target
        # self._max_constraint_val = constraint_value
        self._max_constraint_val = np.inf
        self._constraint_name = constraint_name

        self._opt_fun = LazyDict(
            f_constraint=lambda: tensor_utils.compile_function(
                inputs=inputs + extra_inputs,
                outputs=constraint_term,
                log_name="constraint",
            ),
        )

    def constraint_val(self, inputs, extra_inputs=None):
        """Calculate the constraint value.

        Parameters
        ----------
        inputs :
            A list of symbolic variables as inputs, which could be subsampled if needed. It is assumed
            that the first dimension of these inputs should correspond to the number of data points.
        extra_inputs : optional
            A list of symbolic variables as extra inputs which should not be subsampled.

        Returns
        -------
        constraint_value : float
            The value of the constrained variable.
        """
        inputs = tuple(inputs)
        if extra_inputs is None:
            extra_inputs = tuple()
        return sliced_fun(self._opt_fun["f_constraint"], self._num_slices)(inputs, extra_inputs)

    def get_magnitude(self, direction, inputs, max_constraint_val=None, extra_inputs=None, subsample_grouped_inputs=None):
        """Calculate the update magnitude.

        Parameters
        ----------
        direction: :py:class:'tensorflow.Tensor'
            The gradient direction.
        inputs :
            A list of symbolic variables as inputs, which could be subsampled if needed. It is assumed
            that the first dimension of these inputs should correspond to the number of data points.
        max_constraint_val : float, optional
            The maximum value for the constrained variale.
        extra_inputs : optional
            A list of symbolic variables as extra inputs which should not be subsampled.
        subsample_grouped_inputs : optional
            The list of inputs that are needed to be subsampled.

        Returns
        -------
        magnitude : float
            The update magnitude.
        """
        if max_constraint_val is not None:
            self._max_constraint_val = max_constraint_val
        prev_param = np.copy(self._target.get_param_values(trainable=True))
        inputs = tuple(inputs)
        if extra_inputs is None:
            extra_inputs = tuple()

        if self._subsample_factor < 1:
            if subsample_grouped_inputs is None:
                subsample_grouped_inputs = [inputs]
            subsample_inputs = tuple()
            for inputs_grouped in subsample_grouped_inputs:
                n_samples = len(inputs_grouped[0])
                inds = np.random.choice(
                    n_samples, int(n_samples * self._subsample_factor), replace=False)
                subsample_inputs += tuple([x[inds] for x in inputs_grouped])
        else:
            subsample_inputs = inputs

        Hx = self._hvp_approach.build_eval(subsample_inputs + extra_inputs)

        descent_direction = direction

        initial_step_size = np.sqrt(
            2.0 * self._max_constraint_val * (1. / (descent_direction.dot(Hx(descent_direction)) + 1e-8))
        )
        if np.isnan(initial_step_size):
            initial_step_size = 1.
        flat_descent_step = initial_step_size * descent_direction

        n_iter = 0
        for n_iter, ratio in enumerate(self._backtrack_ratio ** np.arange(self._max_backtracks)):
            cur_step = ratio * flat_descent_step
            cur_param = prev_param - cur_step
            self._target.set_param_values(cur_param, trainable=True)
            constraint_val = sliced_fun(self._opt_fun["f_constraint"], self._num_slices)(inputs, extra_inputs)
            if self._debug_nan and np.isnan(constraint_val):
                import ipdb
                ipdb.set_trace()
            if constraint_val <= self._max_constraint_val:
                break
        if (np.isnan(constraint_val) or constraint_val >= self._max_constraint_val) and not self._accept_violation:
            logger.log("Line search condition violated. Rejecting the step!")
            if np.isnan(constraint_val):
                logger.log("Violated because constraint %s is NaN" % self._constraint_name)
            if constraint_val >= self._max_constraint_val:
                logger.log("Violated because constraint %s is violated" % self._constraint_name)
            self._target.set_param_values(prev_param, trainable=True)
        # logger.log("backtrack iters: %d" % n_iter)
        # logger.log("final magnitude: " + str(-ratio*initial_step_size))
        logger.log("final kl: " + str(constraint_val))
        # logger.log("optimization finished")
        return -ratio * initial_step_size, constraint_val

    def __getstate__(self):
        """Get the internal state.

        Returns
        -------
        data : dict
            The intertal state dict.
        """
        new_dict = self.__dict__.copy()
        del new_dict['_opt_fun']
        return new_dict



import pickle

import numpy as np
import pytest
import tensorflow as tf

from garage.tf.optimizers import PenaltyLbfgsOptimizer
from garage.tf.regressors import GaussianMLPRegressor
from tests.fixtures import TfGraphTestCase


class TestGaussianMLPRegressor(TfGraphTestCase):
    # unmarked to balance test jobs
    # @pytest.mark.large
    def test_fit_normalized(self):
        gmr = GaussianMLPRegressor(input_shape=(1, ), output_dim=1)
        data = np.linspace(-np.pi, np.pi, 1000)
        obs = [{'observations': [[x]], 'returns': [np.sin(x)]} for x in data]

        observations = np.concatenate([p['observations'] for p in obs])
        returns = np.concatenate([p['returns'] for p in obs])
        returns = returns.reshape((-1, 1))
        for _ in range(150):
            gmr.fit(observations, returns)

        paths = {
            'observations': [[-np.pi], [-np.pi / 2], [-np.pi / 4], [0],
                             [np.pi / 4], [np.pi / 2], [np.pi]]
        }

        prediction = gmr.predict(paths['observations'])

        expected = [[0], [-1], [-0.707], [0], [0.707], [1], [0]]
        assert np.allclose(prediction, expected, rtol=0, atol=0.1)

        x_mean = self.sess.run(gmr.model.networks['default'].x_mean)
        x_mean_expected = np.mean(observations, axis=0, keepdims=True)
        x_std = self.sess.run(gmr.model.networks['default'].x_std)
        x_std_expected = np.std(observations, axis=0, keepdims=True)

        assert np.allclose(x_mean, x_mean_expected)
        assert np.allclose(x_std, x_std_expected)

        y_mean = self.sess.run(gmr.model.networks['default'].y_mean)
        y_mean_expected = np.mean(returns, axis=0, keepdims=True)
        y_std = self.sess.run(gmr.model.networks['default'].y_std)
        y_std_expected = np.std(returns, axis=0, keepdims=True)

        assert np.allclose(y_mean, y_mean_expected)
        assert np.allclose(y_std, y_std_expected)

    # unmarked to balance test jobs
    # @pytest.mark.large
    def test_fit_unnormalized(self):
        gmr = GaussianMLPRegressor(input_shape=(1, ),
                                   output_dim=1,
                                   subsample_factor=0.9,
                                   normalize_inputs=False,
                                   normalize_outputs=False)
        data = np.linspace(-np.pi, np.pi, 1000)
        obs = [{'observations': [[x]], 'returns': [np.sin(x)]} for x in data]

        observations = np.concatenate([p['observations'] for p in obs])
        returns = np.concatenate([p['returns'] for p in obs])
        for _ in range(150):
            gmr.fit(observations, returns.reshape((-1, 1)))

        paths = {
            'observations': [[-np.pi], [-np.pi / 2], [-np.pi / 4], [0],
                             [np.pi / 4], [np.pi / 2], [np.pi]]
        }

        prediction = gmr.predict(paths['observations'])

        expected = [[0], [-1], [-0.707], [0], [0.707], [1], [0]]
        assert np.allclose(prediction, expected, rtol=0, atol=0.1)

        x_mean = self.sess.run(gmr.model.networks['default'].x_mean)
        x_mean_expected = np.zeros_like(x_mean)
        x_std = self.sess.run(gmr.model.networks['default'].x_std)
        x_std_expected = np.ones_like(x_std)
        assert np.array_equal(x_mean, x_mean_expected)
        assert np.array_equal(x_std, x_std_expected)

        y_mean = self.sess.run(gmr.model.networks['default'].y_mean)
        y_mean_expected = np.zeros_like(y_mean)
        y_std = self.sess.run(gmr.model.networks['default'].y_std)
        y_std_expected = np.ones_like(y_std)

        assert np.allclose(y_mean, y_mean_expected)
        assert np.allclose(y_std, y_std_expected)

    # unmarked to balance test jobs
    # @pytest.mark.large
    def test_fit_smaller_subsample_factor(self):
        gmr = GaussianMLPRegressor(input_shape=(1, ),
                                   output_dim=1,
                                   subsample_factor=0.9)
        data = np.linspace(-np.pi, np.pi, 1000)
        obs = [{'observations': [[x]], 'returns': [np.sin(x)]} for x in data]

        observations = np.concatenate([p['observations'] for p in obs])
        returns = np.concatenate([p['returns'] for p in obs])
        for _ in range(150):
            gmr.fit(observations, returns.reshape((-1, 1)))

        paths = {
            'observations': [[-np.pi], [-np.pi / 2], [-np.pi / 4], [0],
                             [np.pi / 4], [np.pi / 2], [np.pi]]
        }

        prediction = gmr.predict(paths['observations'])

        expected = [[0], [-1], [-0.707], [0], [0.707], [1], [0]]
        assert np.allclose(prediction, expected, rtol=0, atol=0.1)

    # unmarked to balance test jobs
    # @pytest.mark.large
    def test_fit_without_trusted_region(self):
        gmr = GaussianMLPRegressor(input_shape=(1, ),
                                   output_dim=1,
                                   use_trust_region=False)
        data = np.linspace(-np.pi, np.pi, 1000)
        obs = [{'observations': [[x]], 'returns': [np.sin(x)]} for x in data]

        observations = np.concatenate([p['observations'] for p in obs])
        returns = np.concatenate([p['returns'] for p in obs])
        for _ in range(150):
            gmr.fit(observations, returns.reshape((-1, 1)))

        paths = {
            'observations': [[-np.pi], [-np.pi / 2], [-np.pi / 4], [0],
                             [np.pi / 4], [np.pi / 2], [np.pi]]
        }

        prediction = gmr.predict(paths['observations'])

        expected = [[0], [-1], [-0.707], [0], [0.707], [1], [0]]
        assert np.allclose(prediction, expected, rtol=0, atol=0.1)

    @pytest.mark.parametrize('output_dim, input_shape',
                             [(1, (1, )), (1, (2, )), (2, (3, )), (2, (1, 1)),
                              (3, (2, 2))])
    def test_log_likelihood_sym(self, output_dim, input_shape):
        gmr = GaussianMLPRegressor(input_shape=input_shape,
                                   output_dim=output_dim,
                                   optimizer=PenaltyLbfgsOptimizer,
                                   optimizer_args=dict())

        new_input_var = tf.compat.v1.placeholder(tf.float32,
                                                 shape=(None, ) + input_shape)
        new_ys_var = tf.compat.v1.placeholder(dtype=tf.float32,
                                              name='ys',
                                              shape=(None, output_dim))

        data = np.random.random(size=input_shape)
        label = np.ones(output_dim)

        outputs = gmr.log_likelihood_sym(new_input_var,
                                         new_ys_var,
                                         name='ll_sym')
        ll_from_sym = self.sess.run(outputs,
                                    feed_dict={
                                        new_input_var: [data],
                                        new_ys_var: [label]
                                    })
        mean, log_std = gmr._f_pdists([data])
        ll = gmr.model.networks['default'].dist.log_likelihood(
            [label], dict(mean=mean, log_std=log_std))
        assert np.allclose(ll, ll_from_sym, rtol=0, atol=1e-5)

    def test_is_pickleable(self):
        gmr = GaussianMLPRegressor(input_shape=(1, ), output_dim=1)

        with tf.compat.v1.variable_scope(
                'GaussianMLPRegressor/GaussianMLPRegressorModel', reuse=True):
            bias = tf.compat.v1.get_variable(
                'dist_params/mean_network/hidden_0/bias')
        bias.load(tf.ones_like(bias).eval())

        result1 = gmr.predict(np.ones((1, 1)))
        h = pickle.dumps(gmr)

        with tf.compat.v1.Session(graph=tf.Graph()):
            gmr_pickled = pickle.loads(h)
            result2 = gmr_pickled.predict(np.ones((1, 1)))
            assert np.array_equal(result1, result2)

    def test_is_pickleable2(self):
        gmr = GaussianMLPRegressor(input_shape=(1, ), output_dim=1)

        with tf.compat.v1.variable_scope(
                'GaussianMLPRegressor/GaussianMLPRegressorModel', reuse=True):
            x_mean = tf.compat.v1.get_variable('normalized_vars/x_mean')
        x_mean.load(tf.ones_like(x_mean).eval())
        x1 = gmr.model.networks['default'].x_mean.eval()
        h = pickle.dumps(gmr)
        with tf.compat.v1.Session(graph=tf.Graph()):
            gmr_pickled = pickle.loads(h)
            x2 = gmr_pickled.model.networks['default'].x_mean.eval()
            assert np.array_equal(x1, x2)



import gym.spaces
import numpy as np
import pytest

from garage.envs.wrappers import Grayscale
from tests.fixtures.envs.dummy import DummyDiscretePixelEnv


class TestGrayscale:
    def setup_method(self):
        self.env = DummyDiscretePixelEnv(random=False)
        self.env_g = Grayscale(DummyDiscretePixelEnv(random=False))

    def teardown_method(self):
        self.env.close()
        self.env_g.close()

    def test_grayscale_invalid_environment_type(self):
        with pytest.raises(ValueError):
            self.env.observation_space = gym.spaces.Discrete(64)
            Grayscale(self.env)

    def test_grayscale_invalid_environment_shape(self):
        with pytest.raises(ValueError):
            self.env.observation_space = gym.spaces.Box(
                low=0, high=255, shape=(4, ), dtype=np.uint8)
            Grayscale(self.env)

    def test_grayscale_observation_space(self):
        assert self.env_g.observation_space.shape == (
            self.env.observation_space.shape[:-1])

    def test_grayscale_reset(self):
        """
        RGB to grayscale conversion using scikit-image.

        Weights used for conversion:
        Y = 0.2125 R + 0.7154 G + 0.0721 B

        Reference:
        http://scikit-image.org/docs/dev/api/skimage.color.html#skimage.color.rgb2grey
        """
        grayscale_output = np.round(
            np.dot(self.env.reset()[:, :, :3],
                   [0.2125, 0.7154, 0.0721])).astype(np.uint8)
        np.testing.assert_array_almost_equal(grayscale_output,
                                             self.env_g.reset())

    def test_grayscale_step(self):
        self.env.reset()
        self.env_g.reset()
        obs, _, _, _ = self.env.step(1)
        obs_g, _, _, _ = self.env_g.step(1)

        grayscale_output = np.round(
            np.dot(obs[:, :, :3], [0.2125, 0.7154, 0.0721])).astype(np.uint8)
        np.testing.assert_array_almost_equal(grayscale_output, obs_g)



"""Action and State Spaces to formulate validation as an AST RL problem"""
from .ast_spaces import ASTSpaces  # noqa
from .example_av_spaces import ExampleAVSpaces  # noqa

__all__ = ['ASTSpaces', 'ExampleAVSpaces']



from examples.AV.example_runner_mcts_av import runner as mcts_runner


def validate_mcts():
    # Overall settings
    max_path_length = 50
    s_0 = [0.0, -4.0, 1.0, 11.17, -35.0]
    base_log_dir = './data'
    # experiment settings
    run_experiment_args = {'snapshot_mode': 'last',
                           'snapshot_gap': 1,
                           'log_dir': None,
                           'exp_name': None,
                           'seed': 0,
                           'n_parallel': 1,
                           'tabular_log_file': 'progress.csv'
                           }

    # runner settings
    runner_args = {'n_epochs': 1,
                   'batch_size': 500,
                   'plot': False
                   }

    # env settings
    env_args = {'id': 'ast_toolbox:GoExploreAST-v1',
                'blackbox_sim_state': True,
                'open_loop': False,
                'fixed_init_state': True,
                's_0': s_0,
                }

    # simulation settings
    sim_args = {'blackbox_sim_state': True,
                'open_loop': False,
                'fixed_initial_state': True,
                'max_path_length': max_path_length
                }

    # reward settings
    reward_args = {'use_heuristic': True}

    # spaces settings
    spaces_args = {}

    sampler_args = {'n_envs': 1,
                    'open_loop': False}

    # MCTS Settings

    mcts_bpq_args = {'N': 10}

    exp_log_dir = base_log_dir
    max_path_length = 50
    s_0 = [0.0, -4.0, 1.0, 11.17, -35.0]
    env_args['s_0'] = s_0
    reward_args['use_heuristic'] = True
    sim_args['max_path_length'] = max_path_length

    # MCTS settings
    run_experiment_args['log_dir'] = exp_log_dir + '/mcts'
    run_experiment_args['exp_name'] = 'mcts'

    for mcts_type in ['mcts', 'mctsbv', 'mctsrs']:
        for stress_test_mode in [1, 2]:
            mcts_algo_args = {'max_path_length': max_path_length,
                              'stress_test_mode': stress_test_mode,
                              'ec': 100.0,
                              'n_itr': 1,
                              'k': 0.5,
                              'alpha': 0.5,
                              'clear_nodes': True,
                              'log_interval': 500,
                              'plot_tree': True,
                              'plot_path': run_experiment_args['log_dir'] + '/' + mcts_type + '_tree',
                              'log_dir': run_experiment_args['log_dir'],
                              }
            mcts_runner(
                mcts_type=mcts_type,
                env_args=env_args,
                run_experiment_args=run_experiment_args,
                sim_args=sim_args,
                reward_args=reward_args,
                spaces_args=spaces_args,
                algo_args=mcts_algo_args,
                bpq_args=mcts_bpq_args,
                runner_args=runner_args,
                sampler_args=sampler_args
            )

    return True


if __name__ == '__main__':
    validate_mcts()



"""An example implementation of an ASTReward for an AV validation scenario."""
import numpy as np  # useful packages for math

from ast_toolbox.rewards import ASTReward  # import base class


# Define the class, inherit from the base
class ExampleAVReward(ASTReward):
    """An example implementation of an ASTReward for an AV validation scenario.

    Parameters
    ----------
    num_peds : int
        The number of pedestrians in the scenario.
    cov_x : float
        Covariance of the x-acceleration.
    cov_y : float
        Covariance of the y-acceleration.
    cov_sensor_noise : float
        Covariance of the sensor noise.
    use_heuristic : bool
        Whether to include a heuristic in the reward based on how close the pedestrian is to the vehicle
        at the end of the trajectory.
    """

    def __init__(self,
                 num_peds=1,
                 cov_x=0.1,
                 cov_y=0.01,
                 cov_sensor_noise=0.1,
                 use_heuristic=True):

        self.c_num_peds = num_peds
        self.c_cov_x = cov_x
        self.c_cov_y = cov_y
        self.c_cov_sensor_noise = cov_sensor_noise
        self.use_heuristic = use_heuristic
        super().__init__()

    def give_reward(self, action, **kwargs):
        """Returns the reward for a given time step.

        Parameters
        ----------
        action : array_like
            Action taken by the AST solver.
        kwargs :
            Accepts relevant info for computing the reward.
        Returns
        -------
        reward : float
            Reward based on the previous action.
        """
        # get the info from the simulator
        info = kwargs['info']
        peds = info["peds"]
        car = info["car"]
        is_goal = info["is_goal"]
        is_terminal = info["is_terminal"]
        dist = peds[:, 2:4] - car[2:4]

        # update reward and done bool

        if (is_goal):  # We found a crash
            reward = 0
        elif (is_terminal):
            # reward = 0
            # Heuristic reward based on distance between car and ped at end
            if self.use_heuristic:
                heuristic_reward = np.min(np.linalg.norm(dist, axis=1))
            else:
                # No Herusitic
                heuristic_reward = 0
            reward = -100000 - 10000 * heuristic_reward  # We reached
            # the horizon with no crash
        else:
            reward = -self.mahalanobis_d(action)  # No crash or horizon yet

        return reward

    def mahalanobis_d(self, action):
        """Calculate the Mahalanobis distance [1]_ between the action and the mean action.

        Parameters
        ----------
        action : array_like
            Action taken by the AST solver.

        Returns
        -------
        float
            The Mahalanobis distance between the action and the mean action.

        References
        ----------
        .. [1] Mahalanobis, Prasanta Chandra. "On the generalized distance in statistics." National Institute of
            Science of India, 1936.
            `<http://library.isical.ac.in:8080/jspui/bitstream/123456789/6765/1/Vol02_1936_1_Art05-pcm.pdf>`_
        """
        # Mean action is 0
        mean = np.zeros((6 * self.c_num_peds, 1))
        # Assemble the diagonal covariance matrix
        cov = np.zeros((self.c_num_peds, 6))
        cov[:, 0:6] = np.array([self.c_cov_x, self.c_cov_y,
                                self.c_cov_sensor_noise, self.c_cov_sensor_noise,
                                self.c_cov_sensor_noise, self.c_cov_sensor_noise])
        big_cov = np.diagflat(cov)

        # subtract the mean from our actions
        dif = np.copy(action)
        dif[::2] -= mean[0, 0]
        dif[1::2] -= mean[1, 0]

        # calculate the Mahalanobis distance
        dist = np.dot(np.dot(dif.T, np.linalg.inv(big_cov)), dif)

        return np.sqrt(dist)



class StateActionStateNode:
    def __init__(self, n, r):
        self.n = n  # UInt64
        self.r = r  # Float64

    def __init__(self):
        self.n = 0
        self.r = 0.0


class StateActionNode:
    def __init__(self, s, n, q):
        self.s = s  # Dict{State,StateActionStateNode}
        self.n = n  # UInt64
        self.q = q  # Float64

    def __init__(self):
        self.s = {}
        self.n = 0
        self.q = 0.0


class StateNode:
    def __init__(self, a, n, v):
        self.a = a  # Dict{Action,StateActionNode}
        self.n = n  # UInt64
        self.v = v

    def __init__(self):
        self.a = {}
        self.n = 0
        self.v = 0.0



import uuid

import pydot


def get_root(tree):
    """Get the root node of the tree.

    Parameters
    ----------
    tree : dict
        The tree.

    Returns
    ----------
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The root state.
    """
    for s in tree.keys():
        if s.parent is None:
            return s


def s2node(s, tree):
    """Transfer the AST state to pydot node.

    Parameters
    ----------
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The AST state.
    tree : dict
        The tree.

    Returns
    ----------
    node : :py:class:`pydot.Node`
        The pydot node.
    """
    if s in tree.keys():
        return pydot.Node(str(uuid.uuid4()), label='n=' + str(tree[s].n))
    else:
        return None


def add_children(s, s_node, tree, graph, d):
    """Add successors of s into the graph.

    Parameters
    ----------
    s : :py:class:`ast_toolbox.mcts.AdaptiveStressTesting.ASTState`
        The AST state.
    s_node : :py:class:`pydot.Node`
        The pydot node corresponding to s.
    tree : dict
        The tree.
    graph : :py:class:`pydot.Dot`
        The pydot graph.
    d : int
        The depth.
    """
    if d > 0:
        for a in tree[s].a.keys():
            n = tree[s].a[a].n
            q = tree[s].a[a].q
            assert len(tree[s].a[a].s.keys()) == 1
            for ns in tree[s].a[a].s.keys():
                ns_node = s2node(ns, tree)
                if ns_node is not None:
                    graph.add_node(ns_node)
                    graph.add_edge(pydot.Edge(s_node, ns_node, label="n=" + str(n) + " a=" + str(a.get()) + " q=" + str(q)))
                    # graph.add_edge(pydot.Edge(s_node, ns_node))
                    add_children(ns, ns_node, tree, graph, d - 1)


def plot_tree(tree, d, path, format="svg"):
    """Plot the tree.

    Parameters
    ----------
    tree : dict
        The tree.
    d : int
        The depth.
    path : str
        The plotting path.
    format : str
        The plotting format.
    """
    graph = pydot.Dot(graph_type='digraph')
    root = get_root(tree)
    root_node = s2node(root, tree)
    graph.add_node(root_node)
    add_children(root, root_node, tree, graph, d)
    filename = path + "." + format
    if format == "svg":
        graph.write(filename)
    elif format == "png":
        graph.write(filename)



import pickle

from garage.envs.grid_world_env import GridWorldEnv
from tests.helpers import step_env


class TestGridWorldEnv:
    def test_pickleable(self):
        env = GridWorldEnv(desc='8x8')
        round_trip = pickle.loads(pickle.dumps(env))
        assert round_trip
        assert round_trip.start_state == env.start_state
        step_env(round_trip)
        round_trip.close()
        env.close()

    def test_does_not_modify_action(self):
        env = GridWorldEnv(desc='8x8')
        a = env.action_space.sample()
        a_copy = a
        env.reset()
        env.step(a)
        assert a == a_copy
        env.close()



"""Utiliy functions for tensors."""
import gym.spaces
import numpy as np
import scipy.signal


def discount_cumsum(x, discount):
    """Discounted cumulative sum.

    See https://docs.scipy.org/doc/scipy/reference/tutorial/signal.html#difference-equation-filtering  # noqa: E501
    Here, we have y[t] - discount*y[t+1] = x[t]
    or rev(y)[t] - discount*rev(y)[t-1] = rev(x)[t]

    Args:
        x (np.ndarrary): Input.
        discount (float): Discount factor.

    Returns:
        float: Discounted cumulative sum.

    """
    return scipy.signal.lfilter([1], [1, float(-discount)], x[::-1],
                                axis=0)[::-1]


def explained_variance_1d(ypred, y):
    """Explained variation for 1D inputs.

    It is the proportion of the variance in one variable that is explained or
    predicted from another variable.

    Args:
        ypred (np.ndarray): Sample data from the first variable.
        y (np.ndarray): Sample data from the second variable.

    Returns:
        float: The explained variance.

    """
    assert y.ndim == 1 and ypred.ndim == 1
    vary = np.var(y)
    if np.isclose(vary, 0):
        if np.var(ypred) > 0:
            return 0
        return 1
    return 1 - np.var(y - ypred) / (vary + 1e-8)


def flatten_tensors(tensors):
    """Flatten a list of tensors.

    Args:
        tensors (list[numpy.ndarray]): List of tensors to be flattened.

    Returns:
        numpy.ndarray: Flattened tensors.

    """
    if tensors:
        return np.concatenate([np.reshape(x, [-1]) for x in tensors])
    return np.asarray([])


def unflatten_tensors(flattened, tensor_shapes):
    """Unflatten a flattened tensors into a list of tensors.

    Args:
        flattened (numpy.ndarray): Flattened tensors.
        tensor_shapes (tuple): Tensor shapes.

    Returns:
        list[numpy.ndarray]: Unflattened list of tensors.

    """
    tensor_sizes = list(map(np.prod, tensor_shapes))
    indices = np.cumsum(tensor_sizes)[:-1]
    return [
        np.reshape(pair[0], pair[1])
        for pair in zip(np.split(flattened, indices), tensor_shapes)
    ]


def pad_tensor(x, max_len, mode='zero'):
    """Pad tensors.

    Args:
        x (numpy.ndarray): Tensors to be padded.
        max_len (int): Maximum length.
        mode (str): If 'last', pad with the last element, otherwise pad with 0.

    Returns:
        numpy.ndarray: Padded tensor.

    """
    padding = np.zeros_like(x[0])
    if mode == 'last':
        padding = x[-1]
    return np.concatenate(
        [x, np.tile(padding, (max_len - len(x), ) + (1, ) * np.ndim(x[0]))])


def pad_tensor_n(xs, max_len):
    """Pad array of tensors.

    Args:
        xs (numpy.ndarray): Tensors to be padded.
        max_len (int): Maximum length.

    Returns:
        numpy.ndarray: Padded tensor.

    """
    ret = np.zeros((len(xs), max_len) + xs[0].shape[1:], dtype=xs[0].dtype)
    for idx, x in enumerate(xs):
        ret[idx][:len(x)] = x
    return ret


def pad_tensor_dict(tensor_dict, max_len, mode='zero'):
    """Pad dictionary of tensors.

    Args:
        tensor_dict (dict[numpy.ndarray]): Tensors to be padded.
        max_len (int): Maximum length.
        mode (str): If 'last', pad with the last element, otherwise pad with 0.

    Returns:
        dict[numpy.ndarray]: Padded tensor.

    """
    keys = list(tensor_dict.keys())
    ret = dict()
    for k in keys:
        if isinstance(tensor_dict[k], dict):
            ret[k] = pad_tensor_dict(tensor_dict[k], max_len, mode=mode)
        else:
            ret[k] = pad_tensor(tensor_dict[k], max_len, mode=mode)
    return ret


def stack_tensor_dict_list(tensor_dict_list):
    """Stack a list of dictionaries of {tensors or dictionary of tensors}.

    Args:
        tensor_dict_list (dict[list]): a list of dictionaries of {tensors or
            dictionary of tensors}.

    Return:
        dict: a dictionary of {stacked tensors or dictionary of
            stacked tensors}

    """
    keys = list(tensor_dict_list[0].keys())
    ret = dict()
    for k in keys:
        example = tensor_dict_list[0][k]
        dict_list = [x[k] if k in x else [] for x in tensor_dict_list]
        if isinstance(example, dict):
            v = stack_tensor_dict_list(dict_list)
        else:
            v = np.array(dict_list)
        ret[k] = v
    return ret


def concat_tensor_dict_list(tensor_dict_list):
    """Concatenate dictionary of list of tensor.

    Args:
        tensor_dict_list (dict[list]): a list of dictionaries of {tensors or
            dictionary of tensors}.

    Return:
        dict: a dictionary of {stacked tensors or dictionary of
            stacked tensors}

    """
    keys = list(tensor_dict_list[0].keys())
    ret = dict()
    for k in keys:
        example = tensor_dict_list[0][k]
        dict_list = [x[k] if k in x else [] for x in tensor_dict_list]
        if isinstance(example, dict):
            v = concat_tensor_dict_list(dict_list)
        else:
            v = np.concatenate(dict_list, axis=0)
        ret[k] = v
    return ret


def split_tensor_dict_list(tensor_dict):
    """Split dictionary of list of tensor.

    Args:
        tensor_dict (dict[numpy.ndarray]): a dictionary of {tensors or
            dictionary of tensors}.

    Return:
        dict: a dictionary of {stacked tensors or dictionary of
            stacked tensors}

    """
    keys = list(tensor_dict.keys())
    ret = None
    for k in keys:
        vals = tensor_dict[k]
        if isinstance(vals, dict):
            vals = split_tensor_dict_list(vals)
        if ret is None:
            ret = [{k: v} for v in vals]
        else:
            for v, cur_dict in zip(vals, ret):
                cur_dict[k] = v
    return ret


def truncate_tensor_dict(tensor_dict, truncated_len):
    """Truncate dictionary of list of tensor.

    Args:
        tensor_dict (dict[numpy.ndarray]): a dictionary of {tensors or
            dictionary of tensors}.
        truncated_len (int): Length to truncate.

    Return:
        dict: a dictionary of {stacked tensors or dictionary of
            stacked tensors}

    """
    ret = dict()
    for k, v in tensor_dict.items():
        if isinstance(v, dict):
            ret[k] = truncate_tensor_dict(v, truncated_len)
        else:
            ret[k] = v[:truncated_len]
    return ret


def normalize_pixel_batch(env_spec, observations):
    """Normalize the observations (images).

    If the input are images, it normalized into range [0, 1].

    Args:
        env_spec (garage.envs.EnvSpec): Environment specification.
        observations (numpy.ndarray): Observations from environment.

    Returns:
        numpy.ndarray: Normalized observations.

    """
    if isinstance(env_spec.observation_space, gym.spaces.Box):
        if len(env_spec.observation_space.shape) == 3:
            return [obs.astype(np.float32) / 255.0 for obs in observations]
    return observations



import gym
import numpy as np

from tests.fixtures.envs.dummy import DummyEnv


class LazyFrames(object):
    def __init__(self, frames):
        """
        LazyFrames class from baselines.

        Openai baselines use this class for FrameStack environment
        wrapper. It is used for testing garage.envs.wrappers.AtariEnv.
        garge.envs.wrapper.AtariEnv is used when algorithms are trained
        using baselines wrappers, e.g. during benchmarking.
        """
        self._frames = frames
        self._out = None

    def _force(self):
        if self._out is None:
            self._out = np.concatenate(self._frames, axis=-1)
            self._frames = None
        return self._out

    def __array__(self, dtype=None):
        out = self._force()
        if dtype is not None:
            out = out.astype(dtype)
        return out


class DummyDiscretePixelEnvBaselines(DummyEnv):
    """
    A dummy discrete pixel environment.

    This environment is for testing garge.envs.wrapper.AtariEnv.
    """

    def __init__(self):
        super().__init__(random=False, obs_dim=(10, 10, 3), action_dim=5)
        self._observation_space = gym.spaces.Box(
            low=0, high=255, shape=self._obs_dim, dtype=np.uint8)

    @property
    def observation_space(self):
        """Return an observation space."""
        return self._observation_space

    @property
    def action_space(self):
        """Return an action space."""
        return gym.spaces.Discrete(self._action_dim)

    def step(self, action):
        """gym.Env step function."""
        obs = self.observation_space.sample()
        return LazyFrames([obs]), 0, True, dict()

    def reset(self, **kwargs):
        """gym.Env reset function."""
        obs = np.ones(self._obs_dim, dtype=np.uint8)
        return LazyFrames([obs])



"""MLP in TensorFlow."""

import tensorflow as tf


def mlp(input_var,
        output_dim,
        hidden_sizes,
        name,
        input_var2=None,
        concat_layer=-2,
        hidden_nonlinearity=tf.nn.relu,
        hidden_w_init=tf.glorot_uniform_initializer(),
        hidden_b_init=tf.zeros_initializer(),
        output_nonlinearity=None,
        output_w_init=tf.glorot_uniform_initializer(),
        output_b_init=tf.zeros_initializer(),
        layer_normalization=False):
    """Multi-layer perceptron (MLP).

    It maps real-valued inputs to real-valued outputs.

    Args:
        input_var (tf.Tensor): Input tf.Tensor to the MLP.
        output_dim (int): Dimension of the network output.
        hidden_sizes (list[int]): Output dimension of dense layer(s).
            For example, (32, 32) means this MLP consists of two
            hidden layers, each with 32 hidden units.
        name (str): Network name, also the variable scope.
        input_var2 (tf.Tensor): Second input tf.Tensor to the MLP if input
            needs to be concatenated with a layer in the model.
        concat_layer (int): The index of layers at which to concatenate
            input_var2 with the network. If input_var2 is not supplied, this
            arguments is ignored. The indexing works like standard python list
            indexing. Index of 0 refers to the input layer (input_var) while
            an index of -1 points to the last hidden layer. Default parameter
            points to second layer from the end. If the model has only one
            layer, input_var2 is concatenated with that layer.
        hidden_nonlinearity (callable): Activation function for intermediate
            dense layer(s). It should return a tf.Tensor. Set it to
            None to maintain a linear activation.
        hidden_w_init (callable): Initializer function for the weight
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        hidden_b_init (callable): Initializer function for the bias
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        output_nonlinearity (callable): Activation function for output dense
            layer. It should return a tf.Tensor. Set it to None to
            maintain a linear activation.
        output_w_init (callable): Initializer function for the weight
            of output dense layer(s). The function should return a
            tf.Tensor.
        output_b_init (callable): Initializer function for the bias
            of output dense layer(s). The function should return a
            tf.Tensor.
        layer_normalization (bool): Bool for using layer normalization or not.

    Return:
        The output tf.Tensor of the MLP
    """
    n_layers = len(hidden_sizes) + 1
    _merge_inputs = False

    if input_var2 is not None:
        _merge_inputs = True
        if n_layers > 1:
            _concat_layer = (concat_layer % n_layers + n_layers) % n_layers
        else:
            _concat_layer = 0

    with tf.compat.v1.variable_scope(name):
        l_hid = input_var
        for idx, hidden_size in enumerate(hidden_sizes):
            if _merge_inputs and idx == _concat_layer:
                l_hid = tf.keras.layers.concatenate([l_hid, input_var2])

            l_hid = tf.layers.dense(inputs=l_hid,
                                    units=hidden_size,
                                    activation=hidden_nonlinearity,
                                    kernel_initializer=hidden_w_init,
                                    bias_initializer=hidden_b_init,
                                    name='hidden_{}'.format(idx))
            if layer_normalization:
                l_hid = tf.contrib.layers.layer_norm(l_hid)

        if _merge_inputs and _concat_layer == len(hidden_sizes):
            l_hid = tf.keras.layers.concatenate([l_hid, input_var2])

        l_out = tf.layers.dense(inputs=l_hid,
                                units=output_dim,
                                activation=output_nonlinearity,
                                kernel_initializer=output_w_init,
                                bias_initializer=output_b_init,
                                name='output')
    return l_out



import numpy as np
import pytest
import tensorflow as tf
import torch

import garage.tf.misc.tensor_utils as tf_utils
import garage.torch.algos.loss_function_utils as torch_loss_utils
from tests.fixtures import TfGraphTestCase


def stack(d, arr):
    return np.repeat(np.expand_dims(arr, axis=0), repeats=d, axis=0)


ONES = np.ones((4, 6))
ZEROS = np.zeros((4, 6))
ARRANGE = stack(4, np.arange(6))
PI_DIGITS = stack(4, [3, 1, 4, 1, 5, 9])
E_DIGITS = stack(4, [2, 7, 1, 8, 2, 8])
FIBS = stack(4, [1, 1, 2, 3, 5, 8])


class TestLossFunctionUtils(TfGraphTestCase):
    # yapf: disable
    @pytest.mark.parametrize('gae_lambda, rewards_val, baselines_val', [
        (0.4, ONES, ZEROS),
        (0.8, PI_DIGITS, ARRANGE),
        (1.2, ONES, FIBS),
        (1.7, E_DIGITS, PI_DIGITS),
    ])
    # yapf: enable
    def test_compute_advantages(self, gae_lambda, rewards_val, baselines_val):
        discount = 0.99
        max_len = rewards_val.shape[-1]

        torch_advs = torch_loss_utils.compute_advantages(
            discount, gae_lambda, max_len, torch.Tensor(baselines_val),
            torch.Tensor(rewards_val))

        rewards = tf.compat.v1.placeholder(dtype=tf.float32,
                                           name='reward',
                                           shape=[None, None])
        baselines = tf.compat.v1.placeholder(dtype=tf.float32,
                                             name='baseline',
                                             shape=[None, None])
        adv = tf_utils.compute_advantages(discount, gae_lambda, max_len,
                                          baselines, rewards)
        tf_advs = self.sess.run(adv,
                                feed_dict={
                                    rewards: rewards_val,
                                    baselines: baselines_val,
                                })

        assert np.allclose(torch_advs.numpy(),
                           tf_advs.reshape(torch_advs.shape),
                           atol=1e-5)



import pickle

from examples.AV.example_runner_ba_av import runner as ba_runner
from examples.AV.example_runner_ge_av import runner as go_explore_runner


def validate_ge_ba():
    # Overall settings
    max_path_length = 5
    s_0 = [0.0, -4.0, 1.0, 11.17, -35.0]
    base_log_dir = './data'
    # experiment settings
    run_experiment_args = {'snapshot_mode': 'last',
                           'snapshot_gap': 1,
                           'log_dir': None,
                           'exp_name': None,
                           'seed': 0,
                           'n_parallel': 1,
                           'tabular_log_file': 'progress.csv'
                           }

    # runner settings
    runner_args = {'n_epochs': 1,
                   'batch_size': 500,
                   'plot': False
                   }

    # env settings
    env_args = {'id': 'ast_toolbox:GoExploreAST-v1',
                'blackbox_sim_state': True,
                'open_loop': False,
                'fixed_init_state': True,
                's_0': s_0,
                }

    # simulation settings
    sim_args = {'blackbox_sim_state': True,
                'open_loop': False,
                'fixed_initial_state': True,
                'max_path_length': max_path_length
                }

    # reward settings
    reward_args = {'use_heuristic': True}

    # spaces settings
    spaces_args = {}

    sampler_args = {'n_envs': 1,
                    'open_loop': False}

    # Go-Explore Settings

    ge_algo_args = {'db_filename': None,
                    'max_db_size': 150,
                    'max_path_length': max_path_length,
                    'discount': 0.99,
                    'save_paths_gap': 1,
                    'save_paths_path': None,
                    'overwrite_db': True,
                    }

    ge_baseline_args = {}

    ge_policy_args = {}
    # BA Settings

    ba_algo_args = {'expert_trajectory': None,
                    'max_path_length': max_path_length,
                    'epochs_per_step': 1,
                    'scope': None,
                    'discount': 0.99,
                    'gae_lambda': 1.0,
                    'center_adv': True,
                    'positive_adv': False,
                    'fixed_horizon': False,
                    'pg_loss': 'surrogate_clip',
                    'lr_clip_range': 1.0,
                    'max_kl_step': 1.0,
                    'policy_ent_coeff': 0.0,
                    'use_softplus_entropy': False,
                    'use_neg_logli_entropy': False,
                    'stop_entropy_gradient': False,
                    'entropy_method': 'no_entropy',
                    'name': 'PPO',
                    }

    ba_baseline_args = {}

    ba_policy_args = {'name': 'lstm_policy',
                      'hidden_dim': 64,
                      }

    exp_log_dir = base_log_dir
    env_args['s_0'] = s_0
    reward_args['use_heuristic'] = True
    sim_args['max_path_length'] = max_path_length
    ba_algo_args['max_path_length'] = max_path_length

    # GE settings
    run_experiment_args['log_dir'] = exp_log_dir + '/ge'
    run_experiment_args['exp_name'] = 'ge'
    # run_experiment_args['tabular_log_file'] = 'progress.csv'

    ge_algo_args['db_filename'] = run_experiment_args['log_dir'] + '/cellpool'
    ge_algo_args['save_paths_path'] = run_experiment_args['log_dir']
    ge_algo_args['max_path_length'] = max_path_length

    go_explore_runner(
        env_args=env_args,
        run_experiment_args=run_experiment_args,
        sim_args=sim_args,
        reward_args=reward_args,
        spaces_args=spaces_args,
        policy_args=ge_policy_args,
        baseline_args=ge_baseline_args,
        algo_args=ge_algo_args,
        runner_args=runner_args,
        sampler_args=sampler_args
    )

    with open(run_experiment_args['log_dir'] + '/expert_trajectory.p', 'rb') as f:
        expert_trajectories = pickle.load(f)

    run_experiment_args['log_dir'] = run_experiment_args['log_dir'] + '/ba'
    ba_algo_args['expert_trajectory'] = expert_trajectories[-1]

    ba_runner(
        env_args=env_args,
        run_experiment_args=run_experiment_args,
        sim_args=sim_args,
        reward_args=reward_args,
        spaces_args=spaces_args,
        policy_args=ba_policy_args,
        baseline_args=ba_baseline_args,
        algo_args=ba_algo_args,
        runner_args=runner_args,
        sampler_args=sampler_args
    )

    return True


if __name__ == '__main__':
    validate_ge_ba()



import pickle
from unittest import mock

import numpy as np
import pytest
import tensorflow as tf

from garage.tf.envs import TfEnv
from garage.tf.policies import GaussianMLPPolicy
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyBoxEnv
from tests.fixtures.models import SimpleGaussianMLPModel


class TestGaussianMLPPolicy(TfGraphTestCase):

    @pytest.mark.parametrize('obs_dim, action_dim', [
        ((1, ), (1, )),
        ((1, ), (2, )),
        ((2, ), (2, )),
        ((1, 1), (1, 1)),
        ((1, 1), (2, 2)),
        ((2, 2), (2, 2)),
    ])
    @mock.patch('numpy.random.normal')
    def test_get_action(self, mock_normal, obs_dim, action_dim):
        mock_normal.return_value = 0.5
        env = TfEnv(DummyBoxEnv(obs_dim=obs_dim, action_dim=action_dim))
        with mock.patch(('garage.tf.policies.'
                         'gaussian_mlp_policy.GaussianMLPModel'),
                        new=SimpleGaussianMLPModel):
            policy = GaussianMLPPolicy(env_spec=env.spec)

        env.reset()
        obs, _, _, _ = env.step(1)

        action, prob = policy.get_action(obs)

        expected_action = np.full(action_dim, 0.75)
        expected_mean = np.full(action_dim, 0.5)
        expected_log_std = np.full(action_dim, np.log(0.5))

        assert env.action_space.contains(action)
        assert np.array_equal(action, expected_action)
        assert np.array_equal(prob['mean'], expected_mean)
        assert np.array_equal(prob['log_std'], expected_log_std)

        actions, probs = policy.get_actions([obs, obs, obs])
        for action, mean, log_std in zip(actions, probs['mean'],
                                         probs['log_std']):
            assert env.action_space.contains(action)
            assert np.array_equal(action, expected_action)
            assert np.array_equal(prob['mean'], expected_mean)
            assert np.array_equal(prob['log_std'], expected_log_std)

    @pytest.mark.parametrize('obs_dim, action_dim', [
        ((1, ), (1, )),
        ((1, ), (2, )),
        ((2, ), (2, )),
        ((1, 1), (1, 1)),
        ((1, 1), (2, 2)),
        ((2, 2), (2, 2)),
    ])
    def test_dist_info_sym(self, obs_dim, action_dim):
        env = TfEnv(DummyBoxEnv(obs_dim=obs_dim, action_dim=action_dim))
        with mock.patch(('garage.tf.policies.'
                         'gaussian_mlp_policy.GaussianMLPModel'),
                        new=SimpleGaussianMLPModel):
            policy = GaussianMLPPolicy(env_spec=env.spec)

        env.reset()
        obs, _, _, _ = env.step(1)

        obs_dim = env.spec.observation_space.flat_dim
        obs_ph = tf.compat.v1.placeholder(tf.float32, shape=(None, obs_dim))

        dist1_sym = policy.dist_info_sym(obs_ph, name='p1_sym')

        # flatten output
        expected_mean = [np.full(np.prod(action_dim), 0.5)]
        expected_log_std = [np.full(np.prod(action_dim), np.log(0.5))]

        prob = self.sess.run(dist1_sym, feed_dict={obs_ph: [obs.flatten()]})

        assert np.array_equal(prob['mean'], expected_mean)
        assert np.array_equal(prob['log_std'], expected_log_std)

    @pytest.mark.parametrize('obs_dim, action_dim', [
        ((1, ), (1, )),
        ((1, ), (2, )),
        ((2, ), (2, )),
        ((1, 1), (1, 1)),
        ((1, 1), (2, 2)),
        ((2, 2), (2, 2)),
    ])
    def test_is_pickleable(self, obs_dim, action_dim):
        env = TfEnv(DummyBoxEnv(obs_dim=obs_dim, action_dim=action_dim))
        with mock.patch(('garage.tf.policies.'
                         'gaussian_mlp_policy.GaussianMLPModel'),
                        new=SimpleGaussianMLPModel):
            policy = GaussianMLPPolicy(env_spec=env.spec)

        env.reset()
        obs, _, _, _ = env.step(1)
        obs_dim = env.spec.observation_space.flat_dim

        with tf.compat.v1.variable_scope('GaussianMLPPolicy/GaussianMLPModel',
                                         reuse=True):
            return_var = tf.compat.v1.get_variable('return_var')
        # assign it to all one
        return_var.load(tf.ones_like(return_var).eval())
        output1 = self.sess.run(
            policy.model.outputs[:-1],
            feed_dict={policy.model.input: [obs.flatten()]})

        p = pickle.dumps(policy)
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            policy_pickled = pickle.loads(p)
            output2 = sess.run(
                policy_pickled.model.outputs[:-1],
                feed_dict={policy_pickled.model.input: [obs.flatten()]})
            assert np.array_equal(output1, output2)



import pickle

import numpy as np

from garage.envs import PointEnv
from garage.envs.normalized_env import NormalizedEnv
from tests.helpers import step_env


class TestNormalizedEnv:
    def test_pickleable(self):
        inner_env = PointEnv(goal=(1., 2.))
        env = NormalizedEnv(inner_env, scale_reward=10.)
        round_trip = pickle.loads(pickle.dumps(env))
        assert round_trip
        assert round_trip._scale_reward == env._scale_reward
        assert np.array_equal(round_trip.env._goal, env.env._goal)
        step_env(round_trip)
        round_trip.close()
        env.close()

    def test_does_not_modify_action(self):
        inner_env = PointEnv(goal=(1., 2.))
        env = NormalizedEnv(inner_env, scale_reward=10.)
        a = env.action_space.high + 1.
        a_copy = a
        env.reset()
        env.step(a)
        assert np.array_equal(a, a_copy)
        env.close()



import fire
import numpy as np
# Useful imports
# from example_save_trials import *
import tensorflow as tf
# Import the necessary garage classes
from garage.experiment import run_experiment
from garage.np.baselines.zero_baseline import ZeroBaseline
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies.continuous_mlp_policy import ContinuousMLPPolicy

# Import the AST classes
import ast_toolbox.mcts.BoundedPriorityQueues as BPQ
from ast_toolbox.algos import GA
from ast_toolbox.algos import GASM
from ast_toolbox.envs import ASTEnv
from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.samplers import ASTVectorizedSampler
from ast_toolbox.simulators import ExampleAVSimulator
from ast_toolbox.spaces import ExampleAVSpaces


def runner(
    ga_type=None,
    env_args=None,
    run_experiment_args=None,
    sim_args=None,
    reward_args=None,
    spaces_args=None,
    policy_args=None,
    algo_args=None,
    runner_args=None,
    bpq_args=None,
    sampler_args=None,
    # log_dir='.',
):
    if ga_type is None:
        pass

    if env_args is None:
        env_args = {}

    if run_experiment_args is None:
        run_experiment_args = {}

    if sim_args is None:
        sim_args = {}

    if reward_args is None:
        reward_args = {}

    if spaces_args is None:
        spaces_args = {}

    if policy_args is None:
        policy_args = {}

    if algo_args is None:
        algo_args = {}

    if runner_args is None:
        runner_args = {}

    if sampler_args is None:
        sampler_args = {}

    if bpq_args is None:
        bpq_args = {}

    if 'n_parallel' in run_experiment_args:
        n_parallel = run_experiment_args['n_parallel']
    else:
        n_parallel = 1
        run_experiment_args['n_parallel'] = n_parallel

    if 'max_path_length' in sim_args:
        max_path_length = sim_args['max_path_length']
    else:
        max_path_length = 50
        sim_args['max_path_length'] = max_path_length

    if 'batch_size' in runner_args:
        batch_size = runner_args['batch_size']
    else:
        batch_size = max_path_length * n_parallel
        runner_args['batch_size'] = batch_size

    def run_task(snapshot_config, *_):

        seed = 0
        # top_k = 10
        np.random.seed(seed)

        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        with tf.Session(config=config) as sess:
            with tf.variable_scope('AST', reuse=tf.AUTO_REUSE):

                with LocalTFRunner(
                        snapshot_config=snapshot_config, max_cpus=4, sess=sess) as local_runner:

                    # Instantiate the example classes
                    sim = ExampleAVSimulator(**sim_args)
                    reward_function = ExampleAVReward(**reward_args)
                    spaces = ExampleAVSpaces(**spaces_args)

                    # Create the environment
                    if 'id' in env_args:
                        env_args.pop('id')
                    env = ASTEnv(simulator=sim,
                                 reward_function=reward_function,
                                 spaces=spaces,
                                 **env_args
                                 )
                    env = TfEnv(env)

                    policy = ContinuousMLPPolicy(
                        name='ast_agent',
                        env_spec=env.spec,
                        **policy_args
                    )

                    params = policy.get_params()
                    sess.run(tf.variables_initializer(params))

                    # Instantiate the garage objects
                    baseline = ZeroBaseline(env_spec=env.spec)

                    top_paths = BPQ.BoundedPriorityQueue(**bpq_args)

                    sampler_cls = ASTVectorizedSampler
                    sampler_args['sim'] = sim
                    sampler_args['reward_function'] = reward_function

                    if ga_type == 'ga':
                        print('ga')
                        algo = GA(
                            env_spec=env.spec,
                            policy=policy,
                            baseline=baseline,
                            top_paths=top_paths,
                            **algo_args
                        )
                    elif ga_type == 'gasm':
                        print('gasm')
                        algo = GASM(
                            env_spec=env.spec,
                            policy=policy,
                            baseline=baseline,
                            top_paths=top_paths,
                            **algo_args
                        )
                    else:
                        raise NotImplementedError

                    local_runner.setup(algo=algo,
                                       env=env,
                                       sampler_cls=sampler_cls,
                                       sampler_args=sampler_args)

                    # Run the experiment
                    local_runner.train(**runner_args)

    # from garage.experiment.experiment import AttrDict
    # import os
    # tabular_log_file = os.path.join("./", "test.csv")
    # from dowel import logger, tabular
    # import dowel
    # logger.add_output(dowel.CsvOutput(tabular_log_file))
    # run_task(AttrDict(run_experiment_args))
    run_experiment(
        run_task,
        **run_experiment_args,
    )


if __name__ == '__main__':
    fire.Fire()



"""Toy AV simulator and an example AST simulator wrapper"""
from .toy_av_simulator import ToyAVSimulator  # noqa
from .example_av_simulator import ExampleAVSimulator  # noqa

__all__ = ['ToyAVSimulator', 'ExampleAVSimulator']



"""A toy simulator of a scenario of an AV approaching a crosswalk where some pedestrians are crossing."""
import pdb  # Used for debugging

import numpy as np  # Used for math

# Define the class


class ToyAVSimulator():
    """A toy simulator of a scenario of an AV approaching a crosswalk where some pedestrians are crossing.

    The vehicle runs a modified version of the Intelligent Driver Model [1]_. The vehicle treats the closest
    pedestrian in the road as a car to follow. If no pedestrians are in the road, it attempts to maintain the
    desired speed. Noisy observations of the pedestrian are smoothed through an alpha-beta filter [2]_.

    A collision results if any pedestrian's x-distance and y-distance to the ego vehicle are less than the respective
    `min_dist_x` and `min_dist_y`.

    The origin is centered in the middle of the east/west lane and the north/south crosswalk.
    The positive x proceeds east down the lane, the positive y proceeds north across the crosswalk.

    Parameters
    ----------
    num_peds : int
        The number of pedestrians crossing the street.
    dt : float
        The length (in seconds) of each timestep.
    alpha : float
        The alpha parameter in the tracker's alpha-beta filter [2]_.
    beta : float
        The beta parameter in the tracker's alpha-beta filter [2]_.
    v_des : float
        The desired velocity, in meters per second,  for the ego vehicle to maintain
    delta : float
        The delta parameter in the IDM algorithm [1]_.
    t_headway : float
        The headway parameter in the IDM algorithm [1]_.
    a_max : float
        The maximum acceleration parameter in the IDM algorithm [1]_.
    s_min : float
        The minimum follow distance parameter in the IDM algorithm [1]_.
    d_cmf : float
        The maximum comfortable deceleration parameter in the IDM algorithm [1]_.
    d_max : float
        The maximum deceleration parameter in the IDM algorithm [1]_.
    min_dist_x : float
        The minimum x-distance between the ego vehicle and a pedestrian.
    min_dist_y : float
        The minimum y-distance between the ego vehicle and a pedestrian.
    car_init_x : float
        The initial x-position of the ego vehicle.
    car_init_y : float
        The initial y-position of the ego vehicle.

    References
    ----------
    .. [1] Treiber, Martin, Ansgar Hennecke, and Dirk Helbing.
        "Congested traffic states in empirical observations and microscopic simulations."
        Physical review E 62.2 (2000): 1805.
        `<https://journals.aps.org/pre/abstract/10.1103/PhysRevE.62.1805>`_
    .. [2] Rogers, Steven R. "Alpha-beta filter with correlated measurement noise."
        IEEE Transactions on Aerospace and Electronic Systems 4 (1987): 592-594.
        `<https://ieeexplore.ieee.org/abstract/document/4104388>`_
    """
    # Accept parameters for defining the behavior of the system under test[SUT]

    def __init__(self,
                 num_peds=1,
                 dt=0.1,
                 alpha=0.85,
                 beta=0.005,
                 v_des=11.17,
                 delta=4.0,
                 t_headway=1.5,
                 a_max=3.0,
                 s_min=4.0,
                 d_cmf=2.0,
                 d_max=9.0,
                 min_dist_x=2.5,
                 min_dist_y=1.4,
                 car_init_x=-35.0,
                 car_init_y=0.0,
                 ):

        # Constant hyper-params -- set by user
        self.c_num_peds = num_peds
        self.c_dt = dt
        self.c_alpha = alpha
        self.c_beta = beta
        self.c_v_des = v_des
        self.c_delta = delta
        self.c_t_headway = t_headway
        self.c_a_max = a_max
        self.c_s_min = s_min
        self.c_d_cmf = d_cmf
        self.c_d_max = d_max
        self.c_min_dist = np.array([min_dist_x, min_dist_y])
        self.c_car_init_x = car_init_x
        self.c_car_init_y = car_init_y
        # self.blackbox_sim_state = blackbox_sim_state

        # These are set by reset, not the user
        self._car = np.zeros((4))
        self._car_accel = np.zeros((2))
        self._peds = np.zeros((self.c_num_peds, 4))
        self._measurements = np.zeros((self.c_num_peds, 4))
        self._car_obs = np.zeros((self.c_num_peds, 4))
        self._env_obs = np.zeros((self.c_num_peds, 4))
        self._done = False
        self._reward = 0.0
        self._info = []
        self._step = 0
        self._path_length = 0
        # self._action = None
        self._action = np.array([0] * (6 * self.c_num_peds))
        self._first_step = True
        self.directions = np.random.randint(2, size=self.c_num_peds) * 2 - 1
        self.y = np.random.rand(self.c_num_peds) * 14 - 5
        self.x = np.random.rand(self.c_num_peds) * 4 - 2
        self._state = None

    def run_simulation(self, actions, s_0, simulation_horizon):
        """Run a full simulation given the AST solver's actions and initial conditions.

        Parameters
        ----------
        actions : list[array_like]
            A sequential list of actions taken by the AST Solver which deterministically control the simulation.
        s_0 : array_like
            An array specifying the initial conditions to set the simulator to.
        simulation_horizon : int
            The maximum number of steps a simulation rollout is allowed to run.

        Returns
        -------
        terminal_index : int
            The index of the action that resulted in a state in the goal set E. If no state is found
            terminal_index should be returned as -1.
        array_like
            An array of relevant simulator info, which can then be used for analysis or diagnostics.

        """
        # initialize the simulation
        path_length = 0
        self.reset(s_0)
        self._info = []

        simulation_horizon = np.minimum(simulation_horizon, len(actions))

        # Take simulation steps unbtil horizon is reached
        while path_length < simulation_horizon:
            # get the action from the list
            self._action = actions[path_length]

            # Step the simulation forward in time
            self.step_simulation(self._action)

            # check if a crash has occurred. If so return the timestep, otherwise continue
            if self.collision_detected():
                return path_length, np.array(self._info)
            path_length = path_length + 1

        # horizon reached without crash, return -1
        self._is_terminal = True
        return -1, np.array(self._info)

    def step_simulation(self, action):
        """
        Handle anything that needs to take place at each step, such as a simulation update or write to file.

        Parameters
        ----------
        action : array_like
            A 1-D array of actions taken by the AST Solver which deterministically control
            a single step forward in the simulation.

        Returns
        -------
        array_like
            An observation from the timestep, determined by the settings and the `observation_return` helper function.

        """
        # return None

        # get the action from the list
        self._action = action

        # move the peds
        self.update_peds()

        # move the car
        self._car = self.move_car(self._car, self._car_accel)

        # take new measurements and noise them
        noise = self._action.reshape((self.c_num_peds, 6))[:, 2:6]
        self._measurements = self.sensors(self._peds, noise)

        # filter out the noise with an alpha-beta tracker
        self._car_obs = self.tracker(self._car_obs, self._measurements)

        # select the SUT action for the next timestep
        self._car_accel[0] = self.update_car(self._car_obs, self._car[0])

        # grab simulation state, if interactive
        self.observe()
        self.observation = np.ndarray.flatten(self._env_obs)
        # record step variables
        self.log()

        return self.observation

    def reset(self, s_0):
        """Resets the state of the environment, returning an initial observation.

        Parameters
        ----------
        s_0 : array_like
            The initial conditions to reset the simulator to.

        Returns
        -------
        array_like
            An observation from the timestep, determined by the settings and the `observation_return` helper function.
        """

        # initialize variables
        self._info = []
        self._step = 0
        self._path_length = 0
        self._is_terminal = False
        self.initial_conditions = s_0
        self._action = np.array([0] * (6 * self.c_num_peds))
        self._first_step = True

        # Get v_des if it is sampled from a range
        v_des = self.initial_conditions[3 * self.c_num_peds]

        # initialize SUT location
        car_init_x = self.initial_conditions[3 * self.c_num_peds + 1]
        self._car = np.array([v_des, 0.0, car_init_x, self.c_car_init_y])

        # zero out the first SUT acceleration
        self._car_accel = np.zeros((2))

        # initialize pedestrian locations and velocities
        pos = self.initial_conditions[0:2 * self.c_num_peds]
        self.x = pos[0:self.c_num_peds * 2:2]
        self.y = pos[1:self.c_num_peds * 2:2]
        v_start = self.initial_conditions[2 * self.c_num_peds:3 * self.c_num_peds]
        self._peds[0:self.c_num_peds, 0] = np.zeros((self.c_num_peds))
        self._peds[0:self.c_num_peds, 1] = v_start
        self._peds[0:self.c_num_peds, 2] = self.x
        self._peds[0:self.c_num_peds, 3] = self.y

        # Calculate the relative position measurements
        self._measurements = self._peds
        self._env_obs = self._measurements
        self._car_obs = self._measurements

        # return the initial simulation state
        self.observation = np.ndarray.flatten(self._measurements)
        # self.observation = obs
        return self.observation

    def collision_detected(self):
        """
        Returns whether the current state is in the goal set.

        Checks to see if any pedestrian's position violates both the `min_dist_x` and `min_dist_y` constraints.

        Returns
        -------
        bool
            True if current state is in goal set.
        """
        # calculate the relative distances between the pedestrians and the car
        dist = self._peds[:, 2:4] - self._car[2:4]

        # return True if any relative distance is within the SUT's hitbox and the car is still moving
        if (np.any(np.all(np.less_equal(abs(dist), self.c_min_dist), axis=1)) and
                self._car[0] > 0.5):
            return True

        return False

    def log(self):
        """
        Perform any logging steps.

        """
        # Create a cache of step specific variables for post-simulation analysis
        cache = np.hstack([0.0,  # Dummy, will be filled in with trial # during post processing in save_trials.py
                           self._step,
                           np.ndarray.flatten(self._car),
                           np.ndarray.flatten(self._peds),
                           np.ndarray.flatten(self._action),
                           np.ndarray.flatten(self._car_obs),
                           0.0])
        self._info.append(cache)
        self._step += 1

    def sensors(self, peds, noise):
        """Get a noisy observation of the pedestrians' locations and velocities.

        Parameters
        ----------
        peds : array_like
            Positions and velocities of the pedestrians.
        noise : array_like
            Noise to add to the positions and velocities of the pedestrians.

        Returns
        -------
        array_like
            Noisy observation of the pedestrians' locations and velocities.

        """

        measurements = peds + noise
        return measurements

    def tracker(self, estimate_old, measurements):
        """An alpha-beta filter to smooth noisy observations into an estimate of pedestrian state.

        Parameters
        ----------
        estimate_old : array_like
            The smoothed state estimate from the previous timestep.
        measurements : array_like
            The noisy observation of pedestrian state from the current timestep.

        Returns
        -------
        array_like
            The smoothed state estimate of pedestrian state from the current timestep.
        """
        observation = np.zeros_like(estimate_old)

        observation[:, 0:2] = estimate_old[:, 0:2]
        observation[:, 2:4] = estimate_old[:, 2:4] + self.c_dt * estimate_old[:, 0:2]
        residuals = measurements[:, 2:4] - observation[:, 2:4]

        observation[:, 2:4] += self.c_alpha * residuals
        observation[:, 0:2] += self.c_beta / self.c_dt * residuals

        return observation

    def update_car(self, obs, v_car):
        """Calculate the ego vehicle's acceleration.

        Parameters
        ----------
        obs : array_like
            Smoothed estimate of pedestrian state from the `tracker`.
        v_car : float
            Current velocity of the ego vehicle.

        Returns
        -------
        float
            The acceleration of the ego vehicle.

        """
        cond = np.repeat(np.resize(np.logical_and(obs[:, 3] > -1.5, obs[:, 3] < 4.5), (self.c_num_peds, 1)), 4, axis=1)
        in_road = np.expand_dims(np.extract(cond, obs), axis=0)

        if in_road.size != 0:
            mins = np.argmin(in_road.reshape((-1, 4)), axis=0)
            v_oth = obs[mins[3], 0]
            s_headway = obs[mins[3], 2] - self._car[2]
            s_headway = max(10 ** -6, abs(s_headway)) * np.sign(s_headway)  # avoid div by zero error later

            del_v = v_oth - v_car
            s_des = self.c_s_min + v_car * self.c_t_headway - v_car * del_v / (2 * np.sqrt(self.c_a_max * self.c_d_cmf))
            if self.c_v_des > 0.0:
                v_ratio = v_car / self.c_v_des
            else:
                v_ratio = 1.0

            a = self.c_a_max * (1.0 - v_ratio ** self.c_delta - (s_des / s_headway) ** 2)

        else:
            del_v = self.c_v_des - v_car
            a = del_v

        if np.isnan(a):
            pdb.set_trace()
        # pdb.set_trace()
        return np.clip(a, -self.c_d_max, self.c_a_max)

    def move_car(self, car, accel):
        """Update the ego vehicle's state.

        Parameters
        ----------
        car : array_like
            The ego vehicle's state: [x-velocity, y-velocity, x-position, y-position].
        accel : float
            The ago vehicle's acceleration.

        Returns
        -------
        array_like
            An updated version of the ego vehicle's state.

        """
        car[2:4] += self.c_dt * car[0:2]
        car[0:2] += self.c_dt * accel
        return car

    def update_peds(self):
        """Update the pedestrian's state.

        """
        # Update ped state from actions
        action = self._action.reshape((self.c_num_peds, 6))[:, 0:2]

        mod_a = np.hstack((action,
                           self._peds[:, 0:2] + 0.5 * self.c_dt * action))
        if np.any(np.isnan(mod_a)):
            pdb.set_trace()

        self._peds += self.c_dt * mod_a
        # Enforce max abs(velocity) on pedestrians
        self._peds[:, 0:2] = np.clip(self._peds[:, 0:2], a_min=[-4.5, -4.5], a_max=[4.5, 4.5])
        if np.any(np.isnan(self._peds)):
            pdb.set_trace()

    def observe(self):
        """Get the ground truth state of the pedestrian relative to the ego vehicle.

        """
        self._env_obs = self._peds - self._car

    def get_ground_truth(self):
        """Clones the ground truth simulator state.

        Returns
        -------
        dict
            A dictionary of simulator state variables.
        """
        return {'step': self._step,
                'path_length': self._path_length,
                'is_terminal': self._is_terminal,
                'car': self._car,
                'car_accel': self._car_accel,
                'peds': self._peds,
                'car_obs': self._car_obs,
                'action': self._action,
                'initial_conditions': self.initial_conditions,
                }

    def set_ground_truth(self, in_simulator_state):
        """Sets the simulator state variables.

        Parameters
        ----------
        in_simulator_state : dict
            A dictionary of simulator state variables.
        """
        in_simulator_state.copy()

        self._step = in_simulator_state['step']
        self._path_length = in_simulator_state['path_length']
        self._is_terminal = in_simulator_state['is_terminal']
        self._car = in_simulator_state['car']
        self._car_accel = in_simulator_state['car_accel']
        self._peds = in_simulator_state['peds']
        self._car_obs = in_simulator_state['car_obs']
        self._action = in_simulator_state['action']
        self.initial_conditions = np.array(in_simulator_state['initial_conditions'])

        self.observe()
        self.observation = self._env_obs



"""NormalizedInputMLPModel."""
import numpy as np
import tensorflow as tf

from garage.tf.models import MLPModel


class NormalizedInputMLPModel(MLPModel):
    """
    NormalizedInputMLPModel based on garage.tf.models.Model class.

    This class normalized the inputs and pass the normalized input to a
    MLP model, which can be used to perform linear regression to the outputs.

    Args:
        input_shape (tuple[int]): Input shape of the training data.
        output_dim (int): Output dimension of the model.
        name (str): Model name, also the variable scope.
        hidden_sizes (list[int]): Output dimension of dense layer(s) for
            the MLP for mean. For example, (32, 32) means the MLP consists
            of two hidden layers, each with 32 hidden units.
        hidden_nonlinearity (callable): Activation function for intermediate
            dense layer(s). It should return a tf.Tensor. Set it to
            None to maintain a linear activation.
        hidden_w_init (callable): Initializer function for the weight
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        hidden_b_init (callable): Initializer function for the bias
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        output_nonlinearity (callable): Activation function for output dense
            layer. It should return a tf.Tensor. Set it to None to
            maintain a linear activation.
        output_w_init (callable): Initializer function for the weight
            of output dense layer(s). The function should return a
            tf.Tensor.
        output_b_init (callable): Initializer function for the bias
            of output dense layer(s). The function should return a
            tf.Tensor.
        layer_normalization (bool): Bool for using layer normalization or not.
    """

    def __init__(self,
                 input_shape,
                 output_dim,
                 name='NormalizedInputMLPModel',
                 hidden_sizes=(32, 32),
                 hidden_nonlinearity=tf.nn.relu,
                 hidden_w_init=tf.glorot_uniform_initializer(),
                 hidden_b_init=tf.zeros_initializer(),
                 output_nonlinearity=None,
                 output_w_init=tf.glorot_uniform_initializer(),
                 output_b_init=tf.zeros_initializer(),
                 layer_normalization=False):
        super().__init__(
            output_dim=output_dim,
            name=name,
            hidden_sizes=hidden_sizes,
            hidden_nonlinearity=hidden_nonlinearity,
            hidden_w_init=hidden_w_init,
            hidden_b_init=hidden_b_init,
            output_nonlinearity=output_nonlinearity,
            output_w_init=output_w_init,
            output_b_init=output_b_init,
            layer_normalization=layer_normalization)
        self._input_shape = input_shape

    def network_output_spec(self):
        """Network output spec."""
        return ['y_hat', 'x_mean', 'x_std']

    def _build(self, state_input, name=None):
        with tf.compat.v1.variable_scope('normalized_vars'):
            x_mean_var = tf.compat.v1.get_variable(
                name='x_mean',
                shape=(1, ) + self._input_shape,
                dtype=np.float32,
                initializer=tf.zeros_initializer(),
                trainable=False)
            x_std_var = tf.compat.v1.get_variable(
                name='x_std_var',
                shape=(1, ) + self._input_shape,
                dtype=np.float32,
                initializer=tf.ones_initializer(),
                trainable=False)

        normalized_xs_var = (state_input - x_mean_var) / x_std_var

        y_hat = super()._build(normalized_xs_var)

        return y_hat, x_mean_var, x_std_var



import numpy as np

from garage.envs.wrappers import FireReset
from tests.fixtures.envs.dummy import DummyDiscretePixelEnv


class TestFireReset:
    def test_fire_reset(self):
        env = DummyDiscretePixelEnv()
        env_wrap = FireReset(env)
        obs = env.reset()
        obs_wrap = env_wrap.reset()

        assert np.array_equal(obs, np.ones(env.observation_space.shape))
        assert np.array_equal(obs_wrap, np.full(env.observation_space.shape,
                                                2))

        env_wrap.step(2)
        obs_wrap = env_wrap.reset()  # env will call reset again, after fire
        assert np.array_equal(obs_wrap, np.ones(env.observation_space.shape))



import gym
import pytest

from garage.envs import normalize
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import TNPG
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import GaussianMLPPolicy
from tests.fixtures import snapshot_config, TfGraphTestCase


class TestTNPG(TfGraphTestCase):

    @pytest.mark.large
    def test_tnpg_inverted_pendulum(self):
        """Test TNPG with InvertedPendulum-v2 environment."""
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            env = TfEnv(normalize(gym.make('InvertedPendulum-v2')))

            policy = GaussianMLPPolicy(name='policy',
                                       env_spec=env.spec,
                                       hidden_sizes=(32, 32))

            baseline = LinearFeatureBaseline(env_spec=env.spec)

            algo = TNPG(env_spec=env.spec,
                        policy=policy,
                        baseline=baseline,
                        max_path_length=100,
                        discount=0.99,
                        optimizer_args=dict(reg_coeff=5e-1))

            runner.setup(algo, env)

            last_avg_ret = runner.train(n_epochs=10, batch_size=10000)
            assert last_avg_ret > 15

            env.close()



"""Documented breakages and quirks caused by dependencies."""

# openai/gym environments known to not implement render()
#
# e.g.
# > gym/core.py", line 111, in render
# >     raise NotImplementedError
# > NotImplementedError
#
# Tests calling render() on these should verify they raise NotImplementedError
# ```
# with pytest.raises(NotImplementedError):
#     env.render()
# ```
KNOWN_GYM_RENDER_NOT_IMPLEMENTED = [
    # Please keep alphabetized
    'Blackjack-v0',
    'GuessingGame-v0',
    'HotterColder-v0',
    'NChain-v0',
    'Roulette-v0',
]



from examples.AV.example_runner_drl_av import runner as drl_runner


def validate_drl():
    # Overall settings
    max_path_length = 50
    s_0 = [0.0, -4.0, 1.0, 11.17, -35.0]
    base_log_dir = './data'
    # experiment settings
    run_experiment_args = {'snapshot_mode': 'last',
                           'snapshot_gap': 1,
                           'log_dir': None,
                           'exp_name': None,
                           'seed': 0,
                           'n_parallel': 1,
                           'tabular_log_file': 'progress.csv'
                           }

    # runner settings
    runner_args = {'n_epochs': 1,
                   'batch_size': 500,
                   'plot': False
                   }

    # env settings
    env_args = {'id': 'ast_toolbox:GoExploreAST-v1',
                'blackbox_sim_state': True,
                'open_loop': False,
                'fixed_init_state': True,
                's_0': s_0,
                }

    sampler_args = {'n_envs': 1,
                    'open_loop': False}

    # simulation settings
    sim_args = {'blackbox_sim_state': True,
                'open_loop': False,
                'fixed_initial_state': True,
                'max_path_length': max_path_length
                }

    # reward settings
    reward_args = {'use_heuristic': True}

    # spaces settings
    spaces_args = {}

    # DRL Settings

    drl_policy_args = {'name': 'lstm_policy',
                       'hidden_dim': 64,
                       }

    drl_baseline_args = {}

    drl_algo_args = {'max_path_length': max_path_length,
                     'discount': 0.99,
                     'lr_clip_range': 1.0,
                     'max_kl_step': 1.0,
                     # 'log_dir':None,
                     }

    exp_log_dir = base_log_dir
    max_path_length = 50
    s_0 = [0.0, -4.0, 1.0, 11.17, -35.0]
    env_args['s_0'] = s_0
    reward_args['use_heuristic'] = True
    sim_args['max_path_length'] = max_path_length

    # DRL settings
    run_experiment_args['log_dir'] = exp_log_dir + '/drl'
    run_experiment_args['exp_name'] = 'drl'

    drl_algo_args['max_path_length'] = max_path_length

    drl_runner(
        env_args=env_args,
        run_experiment_args=run_experiment_args,
        sim_args=sim_args,
        reward_args=reward_args,
        spaces_args=spaces_args,
        policy_args=drl_policy_args,
        baseline_args=drl_baseline_args,
        algo_args=drl_algo_args,
        runner_args=runner_args,
        sampler_args=sampler_args
    )

    return True


if __name__ == '__main__':
    validate_drl()



import numpy as np

import ast_toolbox.mcts.AdaptiveStressTesting as AST
import ast_toolbox.mcts.AST_MCTS as AST_MCTS
import ast_toolbox.mcts.ASTSim as ASTSim
import ast_toolbox.mcts.MCTSdpw as MCTSdpw
from ast_toolbox.mcts import tree_plot


class MCTS:
    """Monte Carlo Tress Search (MCTS) with double progressive widening (DPW) [1]_ using the env's action space as its action space.

    Parameters
    ----------
    env : :py:class:`ast_toolbox.envs.go_explore_ast_env.GoExploreASTEnv`.
        The environment.
    max_path_length: int
        The maximum search depth.
    ec : float
        The exploration constant used in UCT equation.
    n_itr: int
            The iteration number, the total numeber of environment call is approximately n_itr*max_path_length*max_path_length.
    k : float
        The constraint parameter used in DPW: |N(s,a)|<=kN(s)^alpha.
    alpha : float
        The constraint parameter used in DPW: |N(s,a)|<=kN(s)^alpha.
    clear_nodes : bool
        Whether to clear redundant nodes in tree.
        Set it to True for saving memoray. Set it to False to better tree plotting.
    log_interval : int
        The log interval in terms of environment calls.
    top_paths : :py:class:`ast_toolbox.mcts.BoundedPriorityQueues`, optional
        The bounded priority queue to store top-rewarded trajectories.
    gamma : float, optional
        The discount factor.
    stress_test_mode : int, optional
        The mode of the tree search. 1 for single tree. 2 for multiple trees.
    log_tabular : bool, optional
        Whether to log the training statistics into a tabular file.
    plot_tree : bool, optional
        Whether to plot the resulting searching tree.
    plot_path : str, optional
        The storing path for the tree plot.
    plot_format : str, optional
        The storing format for the tree plot

    References
    ----------
    .. [1] Lee, Ritchie, et al. "Adaptive stress testing of airborne collision avoidance systems."
     2015 IEEE/AIAA 34th Digital Avionics Systems Conference (DASC). IEEE, 2015.
    """

    def __init__(
        self,
        env,
        max_path_length,
        ec,
        n_itr,
        k,
        alpha,
        clear_nodes,
        log_interval,
        top_paths,
        log_dir,
        gamma=1.0,
        stress_test_mode=2,
        log_tabular=True,
        plot_tree=False,
        plot_path=None,
        plot_format='png'
    ):
        self.env = env
        self.stress_test_mode = stress_test_mode
        self.max_path_length = max_path_length
        self.macts_params = MCTSdpw.DPWParams(max_path_length,
                                              gamma,
                                              ec,
                                              2 * max(n_itr * log_interval // max_path_length ** 2, 1),
                                              k,
                                              alpha,
                                              clear_nodes)
        self.log_interval = log_interval
        self.top_paths = top_paths
        self.log_tabular = log_tabular
        self.plot_tree = plot_tree
        self.plot_path = plot_path
        self.plot_format = plot_format
        self.policy = None
        self.log_dir = log_dir
        self.n_itr = n_itr

    def init(self):
        """Initiate AST internal parameters
        """
        ast_params = AST.ASTParams(self.max_path_length, self.log_interval, self.log_tabular, self.log_dir, self.n_itr)
        self.ast = AST.AdaptiveStressTest(p=ast_params, env=self.env, top_paths=self.top_paths)

    def train(self, runner):
        """Start training.

        Parameters
        ----------
        runner : :py:class:`garage.experiment.LocalRunner <garage:garage.experiment.LocalRunner>`
            ``LocalRunner`` is passed to give algorithm the access to ``runner.step_epochs()``, which provides services
            such as snapshotting and sampler control.
        """
        self.init()
        if self.plot_tree:
            if self.stress_test_mode == 2:
                result, tree = AST_MCTS.stress_test2(self.ast, self.macts_params, self.top_paths, verbose=False, return_tree=True)
            else:
                result, tree = AST_MCTS.stress_test(self.ast, self.macts_params, self.top_paths, verbose=False, return_tree=True)
        else:
            if self.stress_test_mode == 2:
                result = AST_MCTS.stress_test2(self.ast, self.macts_params, self.top_paths, verbose=False, return_tree=False)
            else:
                result = AST_MCTS.stress_test(self.ast, self.macts_params, self.top_paths, verbose=False, return_tree=False)
        self.ast.params.log_tabular = False
        print("checking reward consistance")
        for (action_seq, reward_predict) in result:
            [a.get() for a in action_seq]
            reward, _ = ASTSim.play_sequence(self.ast, action_seq, sleeptime=0.0)
            assert np.isclose(reward_predict, reward)
        print("done")
        if self.plot_tree:
            tree_plot.plot_tree(tree, d=self.max_path_length, path=self.plot_path, format=self.plot_format)



import csv
import os.path

import matplotlib
import numpy as np
from matplotlib import pyplot as plt

matplotlib.rcParams.update({'font.size': 15})

n_trial = 10
top_k = 1
batch_size = 1000
max_step = np.inf
max_reward = np.inf
min_reward = -np.inf

exp_name = 'CartpoleNd'
exp_param = 'D1SUT5'
extra_name = 'MCTS'
prepath = "../" + exp_name + "/Data/AST/Lexington/" + exp_param
plot_path = "../" + exp_name + "/Data/Plot/avgtop" + str(top_k) + "/"
policies = [
    # "TRPOB2000Step0.1","TRPOB2000Step1.0","TRPOB2000Step5.0",\
    # "TRPOB50000Step0.1","TRPOB50000Step1.0","TRPOB50000Step5.0",\
    # "TRPOLSTMB2000Step0.1","TRPOLSTMB2000Step1.0","TRPOLSTMB2000Step5.0",\
    # "TRPOLSTMB50000Step0.1","TRPOLSTMB50000Step1.0","TRPOLSTMB50000Step5.0",\
    "MCTSRSEc10.0K0.3A0.3", "MCTSRSEc10.0K0.5A0.5", "MCTSRSEc10.0K0.8A0.8",\
    "MCTSASEc10.0K0.3A0.3", "MCTSASEc10.0K0.5A0.5", "MCTSASEc10.0K0.8A0.8",\
    "MCTSBVEc10.0K0.3A0.3", "MCTSBVEc10.0K0.5A0.5", "MCTSBVEc10.0K0.8A0.8",\
]
plot_name = exp_name + '_' + exp_param + 'avgtop' + str(top_k) + 'trial' + str(n_trial) + extra_name


plts = []
legends = []
fig = plt.figure(figsize=(10, 10))

for (policy_index, policy) in enumerate(policies):
    print(policy)
    Rewards = []
    min_array_length = np.inf
    for trial in range(n_trial):
        file_path = prepath + '/' + policy + '/' + str(trial) + '/process.csv'
        if os.path.exists(file_path):
            print(trial)
            steps = []
            rewards = []
            with open(file_path) as csv_file:
                if '\0' in open(file_path).read():
                    print("you have null bytes in your input file")
                    csv_reader = csv.reader(x.replace('\0', '') for x in csv_file)
                else:
                    csv_reader = csv.reader(csv_file, delimiter=',')

                for (i, row) in enumerate(csv_reader):
                    if i == 0:
                        entry_dict = {}
                        for index in range(len(row)):
                            entry_dict[row[index]] = index
                    else:
                        # print(row[entry_dict["StepNum"]])
                        if int(row[entry_dict["StepNum"]]) > max_step:
                            break
                        if int(row[entry_dict["StepNum"]]) % batch_size == 0:
                            steps.append(int(row[entry_dict["StepNum"]]))
                            avg_top = 0.0
                            for k in range(top_k):
                                avg_top += np.clip(float(row[entry_dict["reward " + str(k)]]), min_reward, max_reward)
                            avg_top /= top_k
                            rewards.append(avg_top)
            if len(rewards) < min_array_length:
                min_array_length = len(rewards)
            Rewards.append(rewards)
            # print(len(rewards))
            # print(steps[-1])
            # print(min_array_length)
    steps = steps[:min_array_length]
    Rewards = [rewards[:min_array_length] for rewards in Rewards]
    plot, = plt.plot(steps, np.mean(Rewards, 0))
    # plot, = plt.plot(steps,np.mean(np.exp(Rewards),0))
    # plot,_,_ = plt.errorbar(steps,np.mean(Rewards,0),yerr=np.std(Rewards,0)/np.sqrt(n_trial),errorevery=10)
    plts.append(plot)
    legends.append(policy + ' ' + str(np.mean(Rewards, 0)[-1]))

plt.legend(plts, legends)
plt.xlabel('Step Number')
plt.ylabel('Average Top ' + str(top_k) + ' Reward')
fig.savefig(plot_path + plot_name)
plt.close(fig)



import numpy as np

from garage.np.baselines.base import Baseline


class ZeroBaseline(Baseline):

    def __init__(self, env_spec):
        pass

    def get_param_values(self, **kwargs):
        return None

    def set_param_values(self, val, **kwargs):
        pass

    def fit(self, paths):
        pass

    def predict(self, path):
        return np.zeros_like(path['rewards'])

    def predict_n(self, paths):
        return [np.zeros_like(path['rewards']) for path in paths]



import pickle

import numpy as np
import pytest
import torch
from torch import nn

from garage.tf.envs import TfEnv
from garage.torch.q_functions import ContinuousMLPQFunction
from tests.fixtures.envs.dummy import DummyBoxEnv


class TestContinuousNNQFunction:
    # yapf: disable
    @pytest.mark.parametrize('hidden_sizes', [
        (1, ), (2, ), (3, ), (1, 1), (2, 2)])
    # yapf: enable
    def test_forward(self, hidden_sizes):
        env_spec = TfEnv(DummyBoxEnv())
        obs_dim = env_spec.observation_space.flat_dim
        act_dim = env_spec.action_space.flat_dim
        obs = torch.ones(obs_dim, dtype=torch.float32).unsqueeze(0)
        act = torch.ones(act_dim, dtype=torch.float32).unsqueeze(0)

        qf = ContinuousMLPQFunction(env_spec=env_spec,
                                    hidden_nonlinearity=None,
                                    hidden_sizes=hidden_sizes,
                                    hidden_w_init=nn.init.ones_,
                                    output_w_init=nn.init.ones_)

        output = qf(obs, act)
        expected_output = torch.full([1, 1],
                                     fill_value=(obs_dim + act_dim) *
                                     np.prod(hidden_sizes),
                                     dtype=torch.float32)
        assert torch.eq(output, expected_output)

    # yapf: disable
    @pytest.mark.parametrize('batch_size, hidden_sizes', [
        (1, (1, )),
        (3, (2, )),
        (9, (3, )),
        (15, (1, 1)),
        (22, (2, 2)),
    ])
    # yapf: enable
    def test_output_shape(self, batch_size, hidden_sizes):
        env_spec = TfEnv(DummyBoxEnv())
        obs_dim = env_spec.observation_space.flat_dim
        act_dim = env_spec.action_space.flat_dim
        obs = torch.ones(batch_size, obs_dim, dtype=torch.float32)
        act = torch.ones(batch_size, act_dim, dtype=torch.float32)

        qf = ContinuousMLPQFunction(env_spec=env_spec,
                                    hidden_nonlinearity=None,
                                    hidden_sizes=hidden_sizes,
                                    hidden_w_init=nn.init.ones_,
                                    output_w_init=nn.init.ones_)
        output = qf(obs, act)

        assert output.shape == (batch_size, 1)

    # yapf: disable
    @pytest.mark.parametrize('hidden_sizes', [
        (1, ), (2, ), (3, ), (1, 5), (2, 7, 10)])
    # yapf: enable
    def test_is_pickleable(self, hidden_sizes):
        env_spec = TfEnv(DummyBoxEnv())
        obs_dim = env_spec.observation_space.flat_dim
        act_dim = env_spec.action_space.flat_dim
        obs = torch.ones(obs_dim, dtype=torch.float32).unsqueeze(0)
        act = torch.ones(act_dim, dtype=torch.float32).unsqueeze(0)

        qf = ContinuousMLPQFunction(env_spec=env_spec,
                                    hidden_nonlinearity=None,
                                    hidden_sizes=hidden_sizes,
                                    hidden_w_init=nn.init.ones_,
                                    output_w_init=nn.init.ones_)

        output1 = qf(obs, act)

        p = pickle.dumps(qf)
        qf_pickled = pickle.loads(p)
        output2 = qf_pickled(obs, act)

        assert torch.eq(output1, output2)



"""A Discrete QFunction-derived policy.

This policy chooses the action that yields to the largest Q-value.
"""
import akro
import numpy as np
import tensorflow as tf

from garage.tf.policies import Policy


class DiscreteQfDerivedPolicy(Policy):
    """DiscreteQfDerived policy.

    Args:
        env_spec (garage.envs.env_spec.EnvSpec): Environment specification.
        qf (garage.q_functions.QFunction): The q-function used.
        name (str): Name of the policy.
    """

    def __init__(self, env_spec, qf, name='DiscreteQfDerivedPolicy'):
        super().__init__(name, env_spec)

        assert isinstance(env_spec.action_space, akro.Discrete)
        self._env_spec = env_spec
        self._qf = qf

        self._initialize()

    def _initialize(self):
        self._f_qval = tf.compat.v1.get_default_session().make_callable(
            self._qf.q_vals,
            feed_list=[self._qf.model.networks['default'].input])

    @property
    def vectorized(self):
        """Vectorized or not."""
        return True

    def get_action(self, observation):
        """Get action from this policy for the input observation.

        Args:
            observation (numpy.ndarray): Observation from environment.

        Returns:
            Single optimal action from this policy.

        """
        q_vals = self._f_qval([observation])
        opt_action = np.argmax(q_vals)

        return opt_action

    def get_actions(self, observations):
        """Get actions from this policy for the input observations.

        Args:
            observations (numpy.ndarray): Observations from environment.

        Returns:
            Optimal actions from this policy.

        """
        q_vals = self._f_qval(observations)
        opt_actions = np.argmax(q_vals, axis=1)

        return opt_actions

    def __getstate__(self):
        """Object.__getstate__."""
        new_dict = self.__dict__.copy()
        del new_dict['_f_qval']
        return new_dict

    def __setstate__(self, state):
        """Object.__setstate__."""
        self.__dict__.update(state)
        self._initialize()



"""Simulator wrappers to formulate validation as an AST RL problem"""
from .ast_simulator import ASTSimulator  # noqa
from .example_av_simulator import ExampleAVSimulator  # noqa

__all__ = ['ASTSimulator', 'ExampleAVSimulator']



import uuid

import matplotlib as mpl
import pydot
from matplotlib import pyplot as plt

mpl.use('Agg')


def get_root(tree):
    for s in tree.keys():
        if s.parent is None:
            return s


def s2node(s, tree):
    if s in tree.keys():
        # return pydot.Node(str(uuid.uuid4()),label='n='+str(tree[s].n))
        return pydot.Node(str(uuid.uuid4()), label=str(tree[s].v))
    else:
        return None


def add_children(s, s_node, tree, graph, d):
    if d > 0:
        for a in tree[s].a.keys():
            tree[s].a[a].n
            tree[s].a[a].q
            assert len(tree[s].a[a].s.keys()) == 1
            for ns in tree[s].a[a].s.keys():
                ns_node = s2node(ns, tree)
                if ns_node is not None:
                    graph.add_node(ns_node)
                    # graph.add_edge(pydot.Edge(s_node, ns_node, label="n="+str(n)+" q="+str(q)))
                    graph.add_edge(pydot.Edge(s_node, ns_node, label=str(ns.action[0])))
                    add_children(ns, ns_node, tree, graph, d - 1)


def get_node_num_next(s, tree, depths, nodeNums, d):
    d = d + 1
    if (len(depths) <= d) and (len(tree[s].a.keys()) > 0):
        depths.append(d)
        nodeNums.append(0)

    for a in tree[s].a.keys():
        for ns in tree[s].a[a].s.keys():
            nodeNums[d] += 1
            if ns in tree.keys():
                get_node_num_next(ns, tree, depths, nodeNums, d)


def plot_tree(tree, d, path, format="svg"):
    graph = pydot.Dot(graph_type='digraph')
    root = get_root(tree)
    root_node = s2node(root, tree)
    graph.add_node(root_node)
    add_children(root, root_node, tree, graph, d)
    filename = path + "." + format
    if format == "svg":
        graph.write_svg(filename)
    elif format == "png":
        graph.write_png(filename)


def plot_node_num(tree, path, format="svg"):
    root = get_root(tree)
    depths = [0]
    nodeNums = [1]
    d = 0
    get_node_num_next(root, tree, depths, nodeNums, d)
    filename = path + "." + format
    plt.plot(depths, nodeNums)
    plt.xlabel('Depth')
    plt.ylabel('Node Number')
    plt.savefig(filename)



import csv
import os.path

from matplotlib import pyplot as plt

# bug! the xlabel are not aligned
top_k = 10
batch_size = 4000
max_step = 5e6

date = "Lexington"
exps = ["cartpole"]
plolicies = ["MCTS_RS", "MCTS_AS", "MCTS_BV", "RLInter", "RLNonInter", "GAInter", "GANonInter"]
colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w']

for exp in exps:
    success_rates = []
    legends = []
    for (policy_index, policy) in enumerate(plolicies):
        print(policy)
        trial_num = 0.0
        success = 0.0
        file_path = 'Data/AST/' + date + '/' + policy + '/' + 'total_result.csv'
        if os.path.exists(file_path):
            with open(file_path) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                file_empty = True
                for (i, row) in enumerate(csv_reader):
                    file_empty = False
                    print(i)
                    if i != 0:
                        trial_num += 1.0
                        if float(row[1]) > 0.0:
                            success += 1.0
                if file_empty is False:
                    success_rates.append(success / trial_num)
                    legends.append(policy)
    fig = plt.figure(figsize=(20, 20))
    ax = plt.gca()
    plt.scatter(range(len(legends)), success_rates)
    ax.set_xticklabels(legends)
    plt.ylabel('success rate')
    fig.savefig('Data/Plot/' + 'success_rate.pdf')
    plt.close(fig)



"""This modules creates a continuous MLP policy network.

A continuous MLP network can be used as policy method in different RL
algorithms. It accepts an observation of the environment and predicts a
continuous action.
"""
import tensorflow as tf

from garage.tf.models import MLPModel
from garage.tf.policies import Policy


class ContinuousMLPPolicy(Policy):
    """Continuous MLP Policy Network.

    The policy network selects action based on the state of the environment.
    It uses neural nets to fit the function of pi(s).

    Args:
        env_spec (garage.envs.env_spec.EnvSpec): Environment specification.
        name (str): Policy name, also the variable scope.
        hidden_sizes (list[int]): Output dimension of dense layer(s).
            For example, (32, 32) means the MLP of this policy consists of two
            hidden layers, each with 32 hidden units.
        hidden_nonlinearity (callable): Activation function for intermediate
            dense layer(s). It should return a tf.Tensor. Set it to
            None to maintain a linear activation.
        hidden_w_init (callable): Initializer function for the weight
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        hidden_b_init (callable): Initializer function for the bias
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        output_nonlinearity (callable): Activation function for output dense
            layer. It should return a tf.Tensor. Set it to None to
            maintain a linear activation.
        output_w_init (callable): Initializer function for the weight
            of output dense layer(s). The function should return a
            tf.Tensor.
        output_b_init (callable): Initializer function for the bias
            of output dense layer(s). The function should return a
            tf.Tensor.
        input_include_goal (bool): Include goal in the observation or not.
        layer_normalization (bool): Bool for using layer normalization or not.

    """

    def __init__(self,
                 env_spec,
                 name='ContinuousMLPPolicy',
                 hidden_sizes=(64, 64),
                 hidden_nonlinearity=tf.nn.relu,
                 hidden_w_init=tf.glorot_uniform_initializer(),
                 hidden_b_init=tf.zeros_initializer(),
                 output_nonlinearity=tf.nn.tanh,
                 output_w_init=tf.glorot_uniform_initializer(),
                 output_b_init=tf.zeros_initializer(),
                 input_include_goal=False,
                 layer_normalization=False):
        super().__init__(name, env_spec)
        action_dim = env_spec.action_space.flat_dim
        self._hidden_sizes = hidden_sizes
        self._hidden_nonlinearity = hidden_nonlinearity
        self._hidden_w_init = hidden_w_init
        self._hidden_b_init = hidden_b_init
        self._output_nonlinearity = output_nonlinearity
        self._output_w_init = output_w_init
        self._output_b_init = output_b_init
        self._input_include_goal = input_include_goal
        self._layer_normalization = layer_normalization
        if self._input_include_goal:
            self.obs_dim = env_spec.observation_space.flat_dim_with_keys(
                ['observation', 'desired_goal'])
        else:
            self.obs_dim = env_spec.observation_space.flat_dim

        self.model = MLPModel(output_dim=action_dim,
                              name='MLPModel',
                              hidden_sizes=hidden_sizes,
                              hidden_nonlinearity=hidden_nonlinearity,
                              hidden_w_init=hidden_w_init,
                              hidden_b_init=hidden_b_init,
                              output_nonlinearity=output_nonlinearity,
                              output_w_init=output_w_init,
                              output_b_init=output_b_init,
                              layer_normalization=layer_normalization)

        self._initialize()

    def _initialize(self):
        state_input = tf.compat.v1.placeholder(tf.float32,
                                               shape=(None, self.obs_dim))

        with tf.compat.v1.variable_scope(self.name) as vs:
            self._variable_scope = vs
            self.model.build(state_input)

        self._f_prob = tf.compat.v1.get_default_session().make_callable(
            self.model.networks['default'].outputs,
            feed_list=[self.model.networks['default'].input])

    def get_action_sym(self, obs_var, name=None):
        """Symbolic graph of the action.

        Args:
            obs_var (tf.Tensor): Tensor input for symbolic graph.
            name (str): Name for symbolic graph.

        Returns:
            tf.Tensor: symbolic graph of the action.

        """
        with tf.compat.v1.variable_scope(self._variable_scope):
            return self.model.build(obs_var, name=name)

    def get_action(self, observation):
        """Get single action from this policy for the input observation.

        Args:
            observation (numpy.ndarray): Observation from environment.

        Returns:
            numpy.ndarray: Predicted action.
            dict: Empty dict since this policy does not model a distribution.

        """
        action = self._f_prob([observation])
        action = self.action_space.unflatten(action)
        return action, dict()

    def get_actions(self, observations):
        """Get multiple actions from this policy for the input observations.

        Args:
            observations (numpy.ndarray): Observations from environment.

        Returns:
            numpy.ndarray: Predicted actions.
            dict: Empty dict since this policy does not model a distribution.

        """
        actions = self._f_prob(observations)
        actions = self.action_space.unflatten_n(actions)
        return actions, dict()

    def get_regularizable_vars(self):
        """Get regularizable weight variables under the Policy scope.

        Returns:
            list(tf.Variable): List of regularizable variables.

        """
        trainable = self.get_trainable_vars()
        return [
            var for var in trainable
            if 'hidden' in var.name and 'kernel' in var.name
        ]

    @property
    def vectorized(self):
        """Vectorized or not.

        Returns:
            bool: vectorized or not.

        """
        return True

    def clone(self, name):
        """Return a clone of the policy.

        It only copies the configuration of the Q-function,
        not the parameters.

        Args:
            name (str): Name of the newly created policy.

        Returns:
            garage.tf.policies.ContinuousMLPPolicy: Clone of this object

        """
        return self.__class__(name=name,
                              env_spec=self._env_spec,
                              hidden_sizes=self._hidden_sizes,
                              hidden_nonlinearity=self._hidden_nonlinearity,
                              hidden_w_init=self._hidden_w_init,
                              hidden_b_init=self._hidden_b_init,
                              output_nonlinearity=self._output_nonlinearity,
                              output_w_init=self._output_w_init,
                              output_b_init=self._output_b_init,
                              input_include_goal=self._input_include_goal,
                              layer_normalization=self._layer_normalization)

    def __getstate__(self):
        """Object.__getstate__.

        Returns:
            dict: the state to be pickled as the contents for the instance.

        """
        new_dict = super().__getstate__()
        del new_dict['_f_prob']
        return new_dict

    def __setstate__(self, state):
        """Object.__setstate__.

        Args:
            state (dict): unpickled state.

        """
        super().__setstate__(state)
        self._initialize()



"""
This script creates a test that fails when garage.tf.algos.REPS performance is
too low.
"""
import gym
import pytest

from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import REPS
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import CategoricalMLPPolicy
from tests.fixtures import snapshot_config, TfGraphTestCase


class TestREPS(TfGraphTestCase):

    @pytest.mark.large
    def test_reps_cartpole(self):
        """Test REPS with gym Cartpole environment."""
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            env = TfEnv(gym.make('CartPole-v0'))

            policy = CategoricalMLPPolicy(env_spec=env.spec,
                                          hidden_sizes=[32, 32])

            baseline = LinearFeatureBaseline(env_spec=env.spec)

            algo = REPS(env_spec=env.spec,
                        policy=policy,
                        baseline=baseline,
                        max_path_length=100,
                        discount=0.99)

            runner.setup(algo, env)

            last_avg_ret = runner.train(n_epochs=10, batch_size=4000)
            assert last_avg_ret > 5

            env.close()



"""Reinforcement learning algorithms which use NumPy as a numerical backend."""
from garage.np.algos.base import RLAlgorithm
from garage.np.algos.batch_polopt import BatchPolopt
from garage.np.algos.cem import CEM
from garage.np.algos.cma_es import CMAES
from garage.np.algos.nop import NOP
from garage.np.algos.off_policy_rl_algorithm import OffPolicyRLAlgorithm

__all__ = [
    'RLAlgorithm', 'BatchPolopt', 'CEM', 'CMAES', 'NOP', 'OffPolicyRLAlgorithm'
]



"""Utility functions for PyTorch."""
import torch


def np_to_torch(array_dict):
    """
    Convert numpy arrays to PyTorch tensors.

     Args:
        dict (dict): Dictionary of data in numpy arrays.

    Returns:
       Dictionary of data in PyTorch tensors.

    """
    for key, value in array_dict.items():
        array_dict[key] = torch.FloatTensor(value)
    return array_dict


def torch_to_np(value_in):
    """
    Convert PyTorch tensors to numpy arrays.

     Args:
        value_in (tuple): Tuple of data in PyTorch tensors.

    Returns:
       Tuple of data in numpy arrays.

    """
    value_out = []
    for v in value_in:
        value_out.append(v.numpy())
    return tuple(value_out)



"""GaussianMLPRegressorModel."""
import numpy as np
import tensorflow as tf

from garage.tf.models import GaussianMLPModel


class GaussianMLPRegressorModel(GaussianMLPModel):
    """
    GaussianMLPRegressor based on garage.tf.models.Model class.

    This class can be used to perform regression by fitting a Gaussian
    distribution to the outputs.

    Args:
        input_shape (tuple[int]): Input shape of the training data.
        output_dim (int): Output dimension of the model.
        name (str): Model name, also the variable scope.
        hidden_sizes (list[int]): Output dimension of dense layer(s) for
            the MLP for mean. For example, (32, 32) means the MLP consists
            of two hidden layers, each with 32 hidden units.
        hidden_nonlinearity (callable): Activation function for intermediate
            dense layer(s). It should return a tf.Tensor. Set it to
            None to maintain a linear activation.
        hidden_w_init (callable): Initializer function for the weight
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        hidden_b_init (callable): Initializer function for the bias
            of intermediate dense layer(s). The function should return a
            tf.Tensor.
        output_nonlinearity (callable): Activation function for output dense
            layer. It should return a tf.Tensor. Set it to None to
            maintain a linear activation.
        output_w_init (callable): Initializer function for the weight
            of output dense layer(s). The function should return a
            tf.Tensor.
        output_b_init (callable): Initializer function for the bias
            of output dense layer(s). The function should return a
            tf.Tensor.
        learn_std (bool): Is std trainable.
        init_std (float): Initial value for std.
        adaptive_std (bool): Is std a neural network. If False, it will be a
            parameter.
        std_share_network (bool): Boolean for whether mean and std share
            the same network.
        std_hidden_sizes (list[int]): Output dimension of dense layer(s) for
            the MLP for std. For example, (32, 32) means the MLP consists
            of two hidden layers, each with 32 hidden units.
        min_std (float): If not None, the std is at least the value of min_std,
            to avoid numerical issues.
        max_std (float): If not None, the std is at most the value of max_std,
            to avoid numerical issues.
        std_hidden_nonlinearity: Nonlinearity for each hidden layer in
            the std network.
        std_output_nonlinearity: Nonlinearity for output layer in
            the std network.
        std_parametrization (str): How the std should be parametrized. There
            are a few options:
        - exp: the logarithm of the std will be stored, and applied a
            exponential transformation
        - softplus: the std will be computed as log(1+exp(x))
        layer_normalization (bool): Bool for using layer normalization or not.
    """

    def __init__(self,
                 input_shape,
                 output_dim,
                 name='GaussianMLPRegressorModel',
                 **kwargs):
        super().__init__(output_dim=output_dim, name=name, **kwargs)
        self._input_shape = input_shape

    def network_output_spec(self):
        """Network output spec."""
        return [
            'means', 'log_stds', 'std_param', 'normalized_means',
            'normalized_log_stds', 'x_mean', 'x_std', 'y_mean', 'y_std', 'dist'
        ]

    def _build(self, state_input, name=None):
        with tf.compat.v1.variable_scope('normalized_vars'):
            x_mean_var = tf.compat.v1.get_variable(
                name='x_mean',
                shape=(1, ) + self._input_shape,
                dtype=np.float32,
                initializer=tf.zeros_initializer(),
                trainable=False)
            x_std_var = tf.compat.v1.get_variable(
                name='x_std_var',
                shape=(1, ) + self._input_shape,
                dtype=np.float32,
                initializer=tf.ones_initializer(),
                trainable=False)
            y_mean_var = tf.compat.v1.get_variable(
                name='y_mean_var',
                shape=(1, self._output_dim),
                dtype=np.float32,
                initializer=tf.zeros_initializer(),
                trainable=False)
            y_std_var = tf.compat.v1.get_variable(
                name='y_std_var',
                shape=(1, self._output_dim),
                dtype=np.float32,
                initializer=tf.ones_initializer(),
                trainable=False)

        normalized_xs_var = (state_input - x_mean_var) / x_std_var

        normalized_mean, normalized_log_std, std_param, dist = super()._build(
            normalized_xs_var)

        with tf.name_scope('mean_network'):
            means_var = normalized_mean * y_std_var + y_mean_var

        with tf.name_scope('std_network'):
            log_stds_var = normalized_log_std + tf.math.log(y_std_var)

        return (means_var, log_stds_var, std_param, normalized_mean,
                normalized_log_std, x_mean_var, x_std_var, y_mean_var,
                y_std_var, dist)
